"""Provides utilities for kinematics and dynamics parameters retrievals.

This is a standalone file.
pypi dependencies: [numpy, pin]

FEATURES:
    - Retrieve all dynamics and kinematics info for any robot (urdf).
    - All methods become available after a single call to the `update()` method of the class, which
        will compute all required kinematics and dynamics data, which can be retreived using the
        other methods in the class.
    - Handles continuous joints in addition to revolute and prismatic joints.
    - Extends the default pinocchio.RobotWrapper class, so all methods from base class still
        available (if required).

TODO:
    - [x] better way to identify continuous joints is to use pin.model.joints[id].nq and
        pin.model.joints[id].nv values.
        These should be 2 & 1 respectively for continuous joints, 6 & 5 for free-floating, and
        1 & 1 for others.
        NOTE: 'composite joints' can have different nq and nv numbers.
        - [x] also can use joints.shortname() to get joint type directly from this list:
        https://github.com/stack-of-tasks/pinocchio/blob/0caf0ca4d07e63834cdc420c703993662c59e01b/include/pinocchio/multibody/joint/joint-collection.hpp#L24-L67
    - [ ] handle additional joint types.
"""

from typing import List, Tuple, TypeAlias, Sequence, Mapping
import pinocchio
import numpy as np


QuatType: TypeAlias = np.ndarray
"""Numpy array representating quaternion in format [x,y,z,w]"""
Vector3D: TypeAlias = np.ndarray
"""Numpy array representating 3D cartesian vector in format [x,y,z]"""


class PinocchioInterface(pinocchio.RobotWrapper):
    """A pinocchio.RobotWrapper extension to get robot kinematics and dynamics values easily.

    This class cannot deal with passive joints, spherical joints or other composite joints.
    It works fine with regular revolute and prisimatic joints, as well as handles continuous
    joints.
    """

    @staticmethod
    def configuration_handle_continuous_joints(
        q: np.ndarray, continuous_joint_q_ids: Sequence[float]
    ) -> np.ndarray:
        """Static method to update the provided pinocchio q vector to handle continuous joints.

        This is to be used when the original indices are filled with the joint
        position values. These will then be replaced with cos(theta) and sin(theta)
        values in sequence at each of these indices (as required by pinocchio)
        for continuous joints only.

        Args:
            q (np.ndarray): The original values where the values at indices `continuous_joint_q_ids`
                is filled with the actual value of joint position encoder.
            continuous_joint_q_ids (Sequence[float]): The indices in the q vector which are
                continuous joints.

        Returns:
            np.ndarray: Updated q vector with appropriate 2D representation for continuous joints.
        """
        if len(continuous_joint_q_ids) == 0:
            return q

        in_q = q.copy()
        sin_q_ids = np.array(continuous_joint_q_ids) + 1
        continuous_j_vals = q[continuous_joint_q_ids].copy()
        in_q[continuous_joint_q_ids] = np.cos(continuous_j_vals)
        in_q[sin_q_ids] = np.sin(continuous_j_vals)
        return in_q

    @staticmethod
    def recover_generalised_pos_from_pinocchio_configuration(
        q: np.ndarray,
        continuous_joint_q_ids: Sequence[float] = None,
    ) -> np.ndarray:
        """Static method to get the generalised coordinate position vector from pinocchio q vector.

        This function is mainly for handling situations where the robot has continuous joints.
        Pinocchio uses a 2D representation for a single continuous joint, and this function
        recovers single joint position values for such joints.

        Args:
            q (np.ndarray): Configuration vector in the format used by Pinocchio.
            continuous_joint_q_ids (Sequence[float], optional): The joint ids of continuous joints.
                Defaults to [].

        Returns:
            np.ndarray: Output generalised coordinate vector after removing pinocchio's
                continuous joint representation.
        """
        in_q = q.copy()
        if continuous_joint_q_ids is not None and len(continuous_joint_q_ids) > 0:
            sin_ids = np.array(continuous_joint_q_ids) + 1
            in_q[continuous_joint_q_ids] = np.arctan2(q[sin_ids], q[continuous_joint_q_ids])
        return in_q

    def __init__(
        self,
        urdf_filename: str,
        floating_base: bool = True,
        verbose: bool = True,
        ee_names: List[str] = None,
    ):
        """Constructor.
        A pinocchio.RobotWrapper extension to get robot kinematics and dynamics values easily.

        This class cannot deal with passive joints, spherical joints or other composite joints.
        It works fine with regular revolute and prisimatic joints, as well as handles continuous
        joints.

        Args:
            urdf_filename (str): path to urdf file of the robot.
            floating_base (bool, optional): Whether the robot has fixed base or floating base.
                Defaults to True.
            verbose (bool, optional): Verbosity flag. Defaults to True.
            ee_names (List[str], optional): List of end-effector links for the robot.
                Defaults to empty list.
        """
        self.floating_base = floating_base
        if self.floating_base is True:
            self.model = pinocchio.buildModelFromUrdf(
                urdf_filename, pinocchio.JointModelFreeFlyer()
            )
        else:
            self.model = pinocchio.buildModelFromUrdf(urdf_filename)
        super().__init__(self.model, verbose=verbose)

        self.name = self.model.name
        """Name of the robot model."""

        self.total_mass = pinocchio.computeTotalMass(self.model)
        self.num_joints = self.model.njoints
        self.joint_names = self.model.names.tolist()
        self.joint_lower_limits: List[float] = np.array(self.model.lowerPositionLimit)
        self.joint_upper_limits: List[float] = np.array(self.model.upperPositionLimit)
        self.joint_neutral_position = (self.joint_lower_limits + self.joint_upper_limits) / 2
        self.joint_velocity_limit: List[float] = np.array(self.model.velocityLimit)

        self.actuated_joint_names: List[str] = []
        self.actuated_joint_q_ids: List[int] = []
        """index of joints in the generalised coordinate vector (q)"""

        self.q = np.zeros(self.model.nq)  # will be filled with first call to update()
        """The pinocchio q vector (generalised coordinates)."""

        if self.floating_base:
            # set floating joint orientation to unit quaternion
            self.q[7] = 1.0

        self.v = np.zeros(self.model.nv)  # will be filled with first call to update()
        """The pinocchio v vector (generalised velocity)."""
        self.a = np.zeros(self.model.nv)  # will be filled with first call to update()
        """The pinocchio a output vector (generalised acceleration). NOTE: never filled!"""

        self.actuated_joint_v_ids: List[int] = []
        """index of joints in the generalised velocity vector (v)"""

        self.continuous_joint_q_ids: List[int] = []
        """Indices corresponding to continuous joints in the pinocchio configuration vector q.
        In pinocchio, continuous joints are represented using 2 values [cos(theta), sin(theta)].
        This list contains the index to the first (cos) value (add 1 to the index to get the next).
        """

        self.continuous_joint_names: List[str] = []

        def joint_is_continuous(joint: pinocchio.pinocchio_pywrap.JointModel):
            """Check if the pinocchio joint is continuous (as defined in pinocchio)."""
            if (
                joint.nq == 2
                and joint.nv == 1
                # this below check alone should be enough; using nq, nv just to be sure
                and joint.shortname()
                in [
                    "JointModelRUBX",
                    "JointModelRUBY",
                    "JointModelRUBZ",
                    "JointModelRevoluteUnboundedUnaligned",  # unaligned continuous joints
                ]
            ):
                return True
            return False

        for n, name in enumerate(self.joint_names):
            if name not in ["universe", "root_joint"]:
                curr_j_id = self.model.idx_qs[self.model.getJointId(name)]
                curr_v_id = self.model.idx_vs[self.model.getJointId(name)]
                self.actuated_joint_names.append(name)
                self.actuated_joint_q_ids.append(curr_j_id)
                self.actuated_joint_v_ids.append(curr_v_id)
                # To handle continuous joint in the generalised coordinates (q), save these ids
                # separately pinocchio represents continuous joints with 2 numbers (cos(theta),
                # sin(theta))
                if joint_is_continuous(self.model.joints[n]):
                    self.continuous_joint_q_ids.append(curr_j_id)
                    self.continuous_joint_names.append(name)
                # NOTE (@ssidhik): handle composite joints, passive joints, etc.

        self.num_of_actuated_joints = len(self.actuated_joint_names)

        self.actuated_joint_q_ids = np.array(self.actuated_joint_q_ids)
        self.actuated_joint_v_ids = np.array(self.actuated_joint_v_ids)

        additional_dims_ = 7 if self.floating_base else 0
        expected_nq_ = (
            self.num_of_actuated_joints + len(self.continuous_joint_q_ids) + additional_dims_
        )

        assert (
            expected_nq_ == self.nq
        ), f"Could not match expected nq for model. Required model nq: {self.nq}; Current nq obtained: {expected_nq_}."

        self.actuated_joint_name_to_q_index: Mapping[str, int] = dict(
            zip(self.actuated_joint_names, self.actuated_joint_q_ids)
        )
        """Mapping from joint name to index of the joint configuration value in the pinocchio q
        vector."""

        self.actuated_joint_name_to_v_index: Mapping[str, int] = dict(
            zip(self.actuated_joint_names, self.actuated_joint_v_ids)
        )
        """Mapping from joint name to index of the joint velocity value in the pinocchio v
        vector."""

        self.actuated_joint_lower_limits = self.get_joint_positions_from_configuration(
            self.joint_lower_limits
        )
        self.actuated_joint_upper_limits = self.get_joint_positions_from_configuration(
            self.joint_upper_limits
        )
        self.actuated_joint_neutral_position = (
            self.actuated_joint_lower_limits + self.actuated_joint_upper_limits
        ) / 2

        self.actuated_joint_limits = {}
        """mapping from {'joint_name' -> [lower_lim, upper_lim]}"""
        for n, name in enumerate(self.actuated_joint_names):
            self.actuated_joint_limits[name] = [
                self.actuated_joint_lower_limits[n],
                self.actuated_joint_upper_limits[n],
            ]

        self.num_of_frames = self.model.nframes
        self.frame_names = [frame.name for frame in self.model.frames]
        self.frame_ids = [self.model.getFrameId(frame_name) for frame_name in self.frame_names]

        self.ee_names = ee_names if ee_names is not None else []
        """Names of frames that are used as end-effectors."""

        if verbose:
            self.print_model_info()

    def print_model_info(self):
        print("\n")
        print("*" * 100 + "\nPinocchio Model Info " + "\u2193 " * 20 + "\n" + "*" * 100)
        print("name: ", self.model.name)
        print("total_mass:", self.total_mass)
        print("floating base:", self.floating_base)
        print("nq: ", self.nq)
        print("nv: ", self.nv)
        print("joint names:", self.num_joints, self.joint_names)
        print("End-effectors:", self.ee_names)
        print(
            f"Number of moveable joints: {len(self.actuated_joint_q_ids)}, of which "
            f"{len(self.continuous_joint_q_ids)} are continuous joints (with 2 values in pinochio "
            "generalised coordinates)."
        )
        print(
            "actuated joint names: ",
            self.actuated_joint_names,
        )
        print(
            "continuous joint names: ",
            self.continuous_joint_names,
        )
        print("frame names:", self.num_of_frames, self.frame_names)
        print("*" * 100 + "\nPinocchio Model Info " + "\u2191 " * 20 + "\n" + "*" * 100)
        print("\n")

    def update(
        self,
        global_base_position: Vector3D,
        global_base_quaternion: QuatType,
        local_base_velocity_linear: Vector3D,
        local_base_velocity_angular: Vector3D,
        joint_positions: np.ndarray,
        joint_velocities: np.ndarray,
        joint_order: List[str] = None,
    ):
        """Do all forward kinematics and dynamics computations for the given robot states.

        Also updates the internal configuration vectors (self.q, self.v).

        Args:
            global_base_position (Vector3D): Position of robot base in the world.
            global_base_quaternion (QuatType): Orientation quaternion (x,y,z,w) of the
                robot base in the world.
            local_base_velocity_linear (Vector3D): Linear velocity of the robot base in the local
                base frame.
            local_base_velocity_angular (Vector3D): Angular velocity of the robot base in the local
                base frame.
            joint_positions (np.ndarray): Current joint positions of the robot.
            joint_velocities (np.ndarray): Current joint velocities of the robot.
            joint_order (List[str], optional): Joint names in the order the position values were
                given. If None provided, uses default order found by pinocchio (urdf order).
        """
        self.q, self.v = self.get_pinocchio_q_and_v_from_robot_state(
            global_base_position=global_base_position,
            global_base_quaternion=global_base_quaternion,
            local_base_velocity_linear=local_base_velocity_linear,
            local_base_velocity_angular=local_base_velocity_angular,
            joint_positions=joint_positions,
            joint_velocities=joint_velocities,
            joint_order=joint_order,
        )
        self.a = np.zeros(self.model.nv)

        pinocchio.forwardKinematics(self.model, self.data, self.q, self.v, self.a)
        pinocchio.framesForwardKinematics(self.model, self.data, self.q)
        pinocchio.computeJointJacobians(self.model, self.data, self.q)
        pinocchio.computeCentroidalMap(self.model, self.data, self.q)
        pinocchio.computeCentroidalMapTimeVariation(self.model, self.data, self.q, self.v)
        pinocchio.ccrba(self.model, self.data, self.q, self.v)

    def get_frame_id(self, frame_name: str) -> int:
        return self.model.getFrameId(frame_name)

    def get_com_position(self) -> Vector3D:
        return self.com(self.q)

    def get_com_velocity(self) -> Vector3D:
        return self.vcom(self.q, self.v)

    def get_centroidal_mass(self) -> float:
        return self.data.Ig.mass

    def get_gravity_vector(self) -> np.ndarray:
        return self.gravity(q=self.q)

    def get_centroidal_inertia(self) -> np.ndarray:
        return self.data.Ig.inertia

    def get_centroidal_momentum_map(self) -> np.ndarray:
        return self.data.Ag

    def get_centroidal_momentum_map_linear(self) -> np.ndarray:
        return self.get_centroidal_momentum_map()[:3, :]

    def get_centroidal_momentum_map_angular(self) -> np.ndarray:
        return self.get_centroidal_momentum_map()[-3:, :]

    def get_inertia_matrix(self) -> np.ndarray:
        return self.mass(self.q)

    def get_nonlinear_effects(self) -> np.ndarray:
        return self.nle(self.q, self.v)

    def get_joint_positions(self, joint_names: List[str] = None) -> np.ndarray:
        """Get the current joint positions using the latest joint configuration vector.

        This method also handles continuous joints and extracts joint position values for
        continuous joints from the 2D configuration representation used in pinocchio.

        Args:
            joint_names (List[str], optional): Optional list of joints/joint order. Defaults to None
                (uses default joint order).

        Returns:
            np.ndarray: Values of joint position/orientation for specified/all joints.
        """
        return self.get_joint_positions_from_configuration(q=self.q, joint_names=joint_names)

    def get_joint_positions_from_configuration(
        self, q: np.ndarray, joint_names: List[str] = None
    ) -> np.ndarray:
        """Get the joint positions from the provided pinocchio configuration vector.

        This method also handles continuous joints and extracts joint position values for
        continuous joints from the 2D configuration representation used in pinocchio.

        NOTE: This assumes the provided q vector has the same order as the self.q values in this
        class.

        Args:
            q (np.ndarray): Pinocchio configuration vector from which joint values are to be
                extracted. Assumes the provided q vector has the same order as the self.q values in
                this class.
            joint_names (List[str], optional): Optional list of joints/joint order. Defaults to None
                (uses default joint order).

        Returns:
            np.ndarray: Values of joint position/orientation for specified/all joints.
        """
        q = self.get_generalised_coordinates_from_q(q)
        if joint_names is not None and joint_names != self.actuated_joint_names:
            indexes = [self.actuated_joint_name_to_q_index[jname] for jname in joint_names]
            return q[indexes]
        return q[self.actuated_joint_q_ids]

    def get_generalised_coordinates_from_q(self, q: np.ndarray) -> np.ndarray:
        """Get the generalised coordinate representation for the state of the robot.

        This is going to be ordered joint position values in case of fixed-base robots,
        and concatenation of base position, base quaternion (x,y,z,w) and joint positions
        for floating base systems.

        Args:
            q (np.ndarray): Input Pinocchio q vector.

        Returns:
            np.ndarray: output generalised coordinate state representation of this robot.
        """
        assert (
            len(q) == self.nq
        ), f"Input to this function should be a vector of size {self.nq}; got {len(q)}"
        return PinocchioInterface.recover_generalised_pos_from_pinocchio_configuration(
            q=q,
            continuous_joint_q_ids=self.continuous_joint_q_ids,
        )

    def get_pinocchio_q_and_v_from_robot_state(
        self,
        global_base_position: Vector3D,
        global_base_quaternion: QuatType,
        local_base_velocity_linear: Vector3D,
        local_base_velocity_angular: Vector3D,
        joint_positions: np.ndarray,
        joint_velocities: np.ndarray,
        joint_order: List[str] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Get pinocchio configuration vector (q) and velocity (v) from robot state data.

        Also handles continuous joints.

        Args:
            global_base_position (Vector3D): Position of robot base in the world.
            global_base_quaternion (QuatType): Orientation quaternion (x,y,z,w) of the
                robot base in the world.
            local_base_velocity_linear (Vector3D): Linear velocity of the robot base in the local
                base frame.
            local_base_velocity_angular (Vector3D): Angular velocity of the robot base in the local
                base frame.
            joint_positions (np.ndarray): Current joint positions of the robot.
            joint_velocities (np.ndarray): Current joint velocities of the robot.
            joint_order (List[str], optional): Joint names in the order the position values were
                given. If None provided, uses default order found by pinocchio (urdf order).

        Returns:
            Tuple[np.ndarray, np.ndarray]: The q and v vectors that can be used for pinocchio
                operations.
        """
        q = np.zeros(self.model.nq)
        v = np.zeros(self.model.nv)

        if joint_order is not None and joint_order != self.actuated_joint_names:
            assert len(joint_order) == len(self.actuated_joint_names)
            for n, name in enumerate(joint_order):
                q[self.actuated_joint_name_to_q_index[name]] = joint_positions[n]
                v[self.actuated_joint_name_to_v_index[name]] = joint_velocities[n]
        else:
            q[self.actuated_joint_q_ids] = joint_positions
            v[self.actuated_joint_v_ids] = joint_velocities

        if self.floating_base:
            q[:3] = global_base_position
            q[3:7] = global_base_quaternion
            v[:3] = local_base_velocity_linear
            v[3:6] = local_base_velocity_angular

        q = PinocchioInterface.configuration_handle_continuous_joints(
            q=q, continuous_joint_q_ids=self.continuous_joint_q_ids
        )

        return q, v

    def get_frame_position(self, frame_name: str, reference_frame: str = "world") -> Vector3D:
        if reference_frame == "world":
            return self.framePlacement(self.q, self.get_frame_id(frame_name)).translation
        else:
            transform = self.get_frame_transformation(frame_name, reference_frame=reference_frame)
            return transform[:3, 3]

    def get_global_frame_pose(self, frame_name: str) -> Tuple[Vector3D, QuatType]:
        frame_transform = self.framePlacement(self.q, self.get_frame_id(frame_name))
        return (
            frame_transform.translation,
            pinocchio.Quaternion(frame_transform.rotation).coeffs(),
        )

    def get_ee_poses(self, ee_names: List[str] = None) -> List[Tuple[Vector3D, QuatType]]:
        if ee_names is None:
            ee_names = self.ee_names
        return [self.get_global_frame_pose(frame_name=ee_name) for ee_name in ee_names]

    def get_ee_velocities(self, ee_names: List[str] = None) -> List[Tuple[Vector3D, Vector3D]]:
        if ee_names is None:
            ee_names = self.ee_names
        return [
            (
                self.get_frame_velocity_linear(frame_name=ee_name),
                self.get_frame_velocity_angular(frame_name=ee_name),
            )
            for ee_name in ee_names
        ]

    def get_frame_rotation(self, frame_name: str, reference_frame: str = "world") -> np.ndarray:
        if reference_frame == "world":
            return self.framePlacement(self.q, self.get_frame_id(frame_name)).rotation
        else:
            transform = self.get_frame_transformation(frame_name, reference_frame=reference_frame)
            return transform[:3, :3]

    def get_frame_quaternion(self, frame_name: str, reference_frame: str = "world") -> QuatType:
        return pinocchio.Quaternion(self.get_frame_rotation(frame_name, reference_frame)).coeffs()

    def get_frame_transformation(
        self, frame_name: str, reference_frame: str = "world"
    ) -> np.ndarray:
        def homogeneous_matrix(rotation: np.ndarray, translation: np.ndarray) -> np.ndarray:
            mat = np.eye(4)
            mat[:3, :3] = rotation
            mat[:3, 3] = translation
            return mat

        pos = self.get_frame_position(frame_name)
        rot = self.get_frame_rotation(frame_name)
        if reference_frame == "world":
            return homogeneous_matrix(rot, pos)

        transform = homogeneous_matrix(rot, pos)
        pos_reference = self.get_frame_position(reference_frame)
        rot_reference = self.get_frame_rotation(reference_frame)
        transform_reference = homogeneous_matrix(rot_reference, pos_reference)
        local_transformation = np.linalg.pinv(transform_reference) @ transform
        return local_transformation

    def get_frame_velocity_linear(
        self, frame_name: str, reference_frame: str = "world"
    ) -> Vector3D:
        if reference_frame == "world":
            return (
                self.get_frame_rotation(frame_name)
                @ self.frameVelocity(self.q, self.v, self.get_frame_id(frame_name)).linear
            )
        else:
            raise RuntimeError("reference_frame not implemented")

    def get_frame_velocity_angular(
        self, frame_name: str, reference_frame: str = "world"
    ) -> Vector3D:
        if reference_frame == "world":
            return (
                self.get_frame_rotation(frame_name)
                @ self.frameVelocity(self.q, self.v, self.get_frame_id(frame_name)).angular
            )
        else:
            raise RuntimeError("reference_frame not implemented")

    def get_frame_acceleration_linear(
        self, frame_name: str, reference_frame: str = "world"
    ) -> Vector3D:
        if reference_frame == "world":
            return (
                self.get_frame_rotation(frame_name)
                @ self.frameAcceleration(
                    self.q, self.v, self.a, self.get_frame_id(frame_name)
                ).linear
            )
        else:
            raise RuntimeError("reference_frame not implemented")

    def get_frame_acceleration_angular(
        self, frame_name: str, reference_frame: str = "world"
    ) -> Vector3D:
        if reference_frame == "world":
            return (
                self.get_frame_rotation(frame_name)
                @ self.frameAcceleration(
                    self.q, self.v, self.a, self.get_frame_id(frame_name)
                ).angular
            )
        else:
            raise RuntimeError("reference_frame not implemented")

    def get_frame_jacobian(self, frame_name: str, reference_frame: str = "world") -> np.ndarray:
        if reference_frame == "world":
            return self.getFrameJacobian(
                self.get_frame_id(frame_name),
                pinocchio.ReferenceFrame.LOCAL_WORLD_ALIGNED,
            )
        else:
            raise RuntimeError("reference_frame not implemented")

    def get_frame_jacobian_linear(
        self, frame_name: str, reference_frame: str = "world"
    ) -> np.ndarray:
        if reference_frame == "world":
            return self.getFrameJacobian(
                self.get_frame_id(frame_name),
                pinocchio.ReferenceFrame.LOCAL_WORLD_ALIGNED,
            )[:3, :]
        else:
            raise RuntimeError("reference_frame not implemented")

    def get_frame_jacobian_angular(
        self, frame_name: str, reference_frame: str = "world"
    ) -> np.ndarray:
        if reference_frame == "world":
            return self.getFrameJacobian(
                self.get_frame_id(frame_name),
                pinocchio.ReferenceFrame.LOCAL_WORLD_ALIGNED,
            )[-3:, :]
        else:
            raise RuntimeError("reference_frame not implemented")



from typing import TypeAlias, Tuple
import numpy as np
from scipy.spatial.transform import Rotation

QuatType: TypeAlias = np.ndarray
"""Numpy array representating quaternion in format [x,y,z,w]"""
Vector3D: TypeAlias = np.ndarray
"""Numpy array representating 3D cartesian vector in format [x,y,z]"""


def transformation_matrix(
    position: Vector3D = np.array([0, 0, 0]),
    orientation: QuatType = np.array([0, 0, 0, 1]),
) -> np.ndarray:
    """Create a 4x4 transformation matrix

    Args:
        position (Vector3D, optional): position vector. Defaults to np.array([0, 0, 0]).
        orientation (QuatType, optional): orientation quaternion. Defaults to
            np.array([0, 0, 0, 1]).

    Returns:
        np.ndarray: 4x4 transformation matrix representing this pose.
    """
    mat = np.eye(4)
    mat[:3, :3] = quat2rot(quaternion=orientation)
    mat[:3, 3] = position
    return mat


def pos_quat_from_trans_mat(trans_mat: np.ndarray) -> Tuple[Vector3D, QuatType]:
    """Get position and quaternion from a transformation matrix.

    Args:
        trans_mat (np.ndarray): 4x4 transformation matrix.

    Returns:
        Tuple[Vector3D, QuatType]: Equivalent position and quaternion (x,y,z,w)
    """
    return trans_mat[:3, 3], rot2quat(trans_mat[:3, :3])


def invert_transformation_matrix(trans_mat: np.ndarray) -> np.ndarray:
    """Invert the provided homogeneous transformation matrix.

    Args:
        trans_mat (np.ndarray): Input transformation matrix.

    Returns:
        np.ndarray: Inverse of provided transformation matrix.
    """
    rotmat_t = trans_mat[:3, :3].T
    assert is_rotation_matrix(rotmat_t), "Provided matrix is not a valid transformation matrix."
    mat = np.eye(4)
    mat[:3, :3] = rotmat_t
    mat[:3, 3] = -rotmat_t @ trans_mat[:3, 3]
    return mat


def invert_quaternion(quat: QuatType) -> QuatType:
    """Get the quaternion that would give the inverse rotation of the given quaternion.

    Args:
        quat (QuatType): Input quaternion (x,y,z,w)

    Returns:
        QuatType: Inverse of the given quaternion (x,y,z,w)
    """
    out_q = quat.copy()
    out_q[:3] *= -1
    return out_q / np.linalg.norm(out_q)


def quat_error(quat1: QuatType, quat2: QuatType) -> float:
    """Compute shortest angle (in radians) between two quaternions.

    Args:
        quat1 (QuatType): Input quaternion 1. Order: (x,y,z,w)
        quat2 (QuatType): Input quaternion 2. Order: (x,y,z,w)

    Returns:
        float: absoluted error in radians (between 0 and pi)
    """
    return (Rotation.from_quat(quat1) * Rotation.from_quat(quat2).inv()).magnitude()


def quat_diff(quat1: QuatType, quat2: QuatType) -> QuatType:
    """Get the difference quaternion between two quaternions.

    Args:
        quat1 (QuatType): Input quaternion 1 (final). Order: (x,y,z,w)
        quat2 (QuatType): Input quaternion 2 (initial). Order: (x,y,z,w)

    Returns:
        QuatType: Output difference quaternion. Order: (x,y,z,w)
    """
    return (Rotation.from_quat(quat1) * Rotation.from_quat(quat2).inv()).as_quat()


def rpy_error(rpy1: Vector3D, rpy2: QuatType) -> float:
    """Compute shortest angle (in radians) between two rotations (described by roll-pitch-yaw).

    Args:
        rpy1 (QuatType): Input RPY value 1
        rpy2 (QuatType): Input RPY value 2

    Returns:
        float: absoluted error in radians (between 0 and pi)
    """
    return (Rotation.from_euler("xyz", rpy1) * Rotation.from_euler("xyz", rpy2).inv()).magnitude()


def rotmat_error(rotmat1: np.ndarray, rotmat2: np.ndarray):
    """Compute shortest angle (in radians) between two rotations (described by rotation matrices).

    Args:
        rotmat1 (QuatType): Input rotation matrix 1
        rotmat2 (QuatType): Input rotation matrix 2

    Returns:
        float: absoluted error in radians (between 0 and pi)
    """
    return (Rotation.from_matrix(rotmat1) * Rotation.from_matrix(rotmat2).inv()).magnitude()


def wrap_angle(angle: float | np.ndarray) -> float | np.ndarray:
    """Wrap the provided angle(s) (radians) to be within -pi and pi.

    Args:
        angle (float | np.ndarray): Input angle (or array of angles) in radians.

    Returns:
        float | np.ndarray: Output after wrapping.
    """
    return (angle + np.pi) % (2 * np.pi) - np.pi


def quat_multiply(quat1: QuatType, quat2: QuatType) -> QuatType:
    """Multiply quat1 with quat2. (q1 * q2)

    Args:
        quat1 (QuatType): Input quaternion 1
        quat2 (QuatType): Input quaternion 2

    Returns:
        QuatType: Output quaternion (q1*q2)
    """
    return (Rotation.from_quat(quat1) * Rotation.from_quat(quat2)).as_quat()


def quat2rpy(quaternion: QuatType) -> np.ndarray:
    """Convert a quaternion into euler angles (roll, pitch, yaw).

    roll is rotation around x in radians (counterclockwise)
    pitch is rotation around y in radians (counterclockwise)
    yaw is rotation around z in radians (counterclockwise)

    Args:
        quaternion (QuatType): x,y,z,w order

    Returns:
        np.ndarray: Euler engles in rpy order.
    """
    try:
        return Rotation.from_quat(quaternion).as_euler("xyz")
    except ValueError as e:
        raise ValueError(f"{e}: Culprit: {quaternion}") from e


def rpy2quat(rpy: np.ndarray) -> QuatType:
    """Convert an Euler angle to a quaternion.

    Args:
        rpy (np.ndarray): roll, pitch, yaw in radians

    Returns:
        QuatType: The orientation in quaternion [x,y,z,w] format
    """
    return Rotation.from_euler("xyz", rpy).as_quat()


def xyz_local2quat(xyz: np.ndarray) -> QuatType:
    """Convert sequential local x,y,z rotations to a quaternion.

    Args:
        xyz (np.ndarray): orientations along x, y, z (local) in radians

    Returns:
        QuatType: The orientation in quaternion [x,y,z,w] format
    """
    xyz = np.array(xyz).reshape([-1, 3])
    zyx = xyz.copy()
    zyx[:, 0] = xyz[:, 2]
    zyx[:, 2] = xyz[:, 0]
    q = Rotation.from_euler("zyx", zyx).as_quat()
    if q.shape[0] == 1:
        return q.flatten()
    return q


def quat2xyz_local(quat: QuatType) -> np.ndarray:
    """Convert quaternion to sequential local x,y,z rotations.

    Args:
        quat (QuatType): The orientation in quaternion [x,y,z,w] format

    Returns:
        np.ndarray: orientations along x, y, z (local) in radians. Shape: (3,)
    """
    zyx = Rotation.from_quat(quat).as_euler("zyx").reshape([-1, 3])
    xyz = zyx.copy()
    xyz[:, 0] = zyx[:, 2]
    xyz[:, 2] = zyx[:, 0]
    if xyz.shape[0] == 1:
        return xyz.flatten()
    return xyz


def quat2rot(quaternion: QuatType) -> np.ndarray:
    """Covert a quaternion into a full three-dimensional rotation matrix.

    Args:
        quaternion (QuatType): A 4 element array representing the quaternion (w,x,y,z)

    Returns:
        np.ndarray: A 3x3 element matrix representing the full 3D rotation matrix.
             This rotation matrix converts a point in the local reference
             frame to a point in the global reference frame.
    """
    try:
        return Rotation.from_quat(quaternion).as_matrix()
    except ValueError:
        # BUG: occasional nans in quaternion retrieved from pybullet
        # https://github.com/bulletphysics/bullet3/issues/976
        print(f"Quaternion error: {quaternion}")
        raise


def rpy2rot(rpy: np.ndarray) -> np.ndarray:
    """Converts roll, pitch, yaw euler angles to 3x3 rotation matrix.

    Args:
        rpy (np.ndarray): roll, pitch, yaw in radians

    Returns:
        np.ndarray: equivalent 3x3 rotation matrix
    """
    return Rotation.from_euler("xyz", rpy).as_matrix()


def is_rotation_matrix(mat: np.ndarray) -> bool:
    """Checks if a matrix is a valid rotation matrix.

    Args:
        mat (np.ndarray): Input matrix to check

    Returns:
        bool: True if this is a valid rotation matrix.
    """
    return np.linalg.norm(np.eye(3) - np.dot(np.array(mat).T, mat)) < 1e-6


def rot2rpy(rotmat: np.ndarray) -> np.ndarray:
    """Converts 3x3 rotation matrix to euler roll, pitch, yaw angles.

    Args:
        rotmat (np.ndarray): 3x3 rotation matrix.

    Returns:
        np.ndarray: roll, pitch, yaw in radians.
    """
    assert is_rotation_matrix(rotmat), "Provided R matrix is not a valid rotation matrix."
    return Rotation.from_matrix(rotmat).as_euler("xyz")


def rot2quat(rotmat: np.ndarray) -> QuatType:
    """Converts 3x3 rotation matrix to quaternion.

    Args:
        rotmat (np.ndarray): 3x3 rotation matrix.

    Returns:
        QuatType: Quaternion (x,y,z,w)
    """
    assert is_rotation_matrix(rotmat), "Provided R matrix is not a valid rotation matrix."
    return Rotation.from_matrix(rotmat).as_quat()


def vec2skew(vector: Vector3D) -> np.ndarray:
    """Convert provided vector to its equivalent skew-symmetric matrix.

    Args:
        vector (Vector3D): 3d vector.

    Returns:
        np.ndarray: 3x3 skew symmetric matrix.
    """
    skew_mat = np.array(
        [
            [0.0, -vector[2], vector[1]],
            [vector[2], 0.0, -vector[0]],
            [-vector[1], vector[0], 0.0],
        ]
    )

    return skew_mat


def vec2quatskew(vector: Vector3D) -> np.ndarray:
    """Convert provided vector to its "Omega" operator as described in
    https://ahrs.readthedocs.io/en/latest/filters/angular.html.

    Args:
        vector (Vector3D): 3d vector.

    Returns:
        np.ndarray: 4x4 Omega operator matrix.
    """

    omega_matrix = np.array(
        [
            [0, -vector[0], -vector[1], -vector[2]],
            [vector[0], 0, vector[2], -vector[1]],
            [vector[1], -vector[2], 0, vector[0]],
            [vector[2], vector[1], -vector[0], 0],
        ]
    )

    return omega_matrix



from abc import ABC, abstractmethod
from typing import List, Tuple
import copy

from ...core.types import GlobalMotionPlan, LocalMotionPlan, RobotState, RobotCmd
from ...utils.time_utils import RateTrigger, ClockBase, PythonPerfClock
from ...core.logging import logging


class CtrlLoopDebuggerBase(ABC):
    """Base class for control loop debuggers. Can only be used with control loops."""

    def __init__(self, rate: float = None, clock: ClockBase = PythonPerfClock()):
        """Base class for control loop debuggers. Can only be used with control loops.

        Args:
            rate (float, optional): Rate at which this should be triggered. Defaults to None
                (i.e. use control loop rate).
            clock (ClockBase, optional): The clock to use for timer. Defaults to PythonPerfClock().
        """
        self._rate = rate
        if self._rate is not None and self._rate > 0.0:
            self._rate_trigger = RateTrigger(rate=self._rate, clock=clock)
            logging.debug(f"{self.__class__.__name__}: Setting trigger rate to {self._rate}Hz.")

    def _should_run(self):
        if self._rate is None:
            return True
        if self._rate <= 0.0:
            return False
        return self._rate_trigger.triggered()

    def run_once(
        self,
        t: float,
        dt: float,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        agent_outputs: List[Tuple[LocalMotionPlan, RobotCmd]],
        robot_cmd: RobotCmd,
    ):
        """The main method that will be run in every control loop.

        NOTE: SHOULD NOT BE OVERRIDDEN. Override the `_run_once_impl` method.

        This method automatically checks if it is ok to trigger the implemented
        child method, and calls it if appropriate.

        Args:
            t (float): Current time in the control loop.
            dt (float): The current dt in the control loop.
            robot_state (RobotState): robot state after update from the state estimator
                in the control loop.
            global_plan (GlobalMotionPlan): The global plan output produced by the global
                planner.
            agent_outputs (List[Tuple[LocalMotionPlan, RobotCmd]]): The list of outputs
                from all agents in the control loop.
            robot_cmd (RobotCmd): The final command written to the robot.
        """
        if self._should_run():
            self._run_once_impl(
                t=t,
                dt=dt,
                robot_state=copy.deepcopy(robot_state),
                global_plan=copy.deepcopy(global_plan),
                agent_outputs=copy.deepcopy(agent_outputs),
                robot_cmd=copy.deepcopy(robot_cmd),
            )

    @abstractmethod
    def _run_once_impl(
        self,
        t: float,
        dt: float,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        agent_outputs: List[Tuple[LocalMotionPlan, RobotCmd]],
        robot_cmd: RobotCmd,
    ):
        """The main method that will be run in every control loop. Should be overridden
        in the implemented child of `CtrlLoopDebuggerBase`.

        This method will be called automatically if the timer is triggered, as long
        as the main `run_once()` method is called continuously in the control loop.

        Args:
            t (float): Current time in the control loop.
            dt (float): The current dt in the control loop.
            robot_state (RobotState): robot state after update from the state estimator
                in the control loop.
            global_plan (GlobalMotionPlan): The global plan output produced by the global
                planner.
            agent_outputs (List[Tuple[LocalMotionPlan, RobotCmd]]): The list of outputs
                from all agents in the control loop.
            robot_cmd (RobotCmd): The final command written to the robot.
        """
        raise NotImplementedError("This method has to be implemented in the child class")

    def shutdown(self):
        """Cleanly shutdown the debugger. Override in child class if required. The
        base class implements an empty function."""
        return


class DummyDebugger(CtrlLoopDebuggerBase):
    """A dummy debugger to test in pipeline."""

    def __init__(
        self,
        squawk: bool = True,
        rate: float = 10,
        clock: ClockBase = PythonPerfClock(),
    ):
        super().__init__(rate, clock)
        self._squawk = squawk

    def _run_once_impl(
        self,
        t: float,
        dt: float,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        agent_outputs: List[Tuple[LocalMotionPlan | RobotCmd]],
        robot_cmd: RobotCmd,
    ):
        if self._squawk:
            print("Debugger in loop running...")



from typing import Mapping, List, Dict, Callable
import threading
import copy
from functools import partial
import pybullet as pb
import numpy as np
import pybullet_robot

from ...components.robot_interfaces.simulation.pybullet_robot import PybulletRobot
from ..time_utils import RateLimiter
from ..math_utils import rpy2quat, quat2rpy
from ...core.logging import logging
from ...components.callback_handlers.base_callbacks import RateTriggeredMultiCallbacks
from ...components.callback_handlers.pb_gui_callbacks import (
    PbGUIButtonCallback,
    PbMultiGUISliderSingleCallback,
    PbGUICallback,
)

# pylint: disable=W0212,W0201


class PybulletRobotVisualizer:
    """Visualizer for a PybulletRobot instance to test joint positioning etc.

    See `examples/utils_demo/demo_bullet_robot_visualizer.py`.

    WARNING
    -------
    DO NOT use other GUI utils for pybullet (other sliders, buttons etc), when
    using this. This class uses the `removeAllUserParameters` parameters exposed
    in pybullet, which means it can remove every debug item created externally.

    TODO: fix this to only remove self-created items.
    """

    def __init__(
        self,
        pb_sim_interface: pybullet_robot.BulletRobot,
        joint_lower_lims: Mapping[str, float],
        joint_upper_lims: Mapping[str, float],
        starting_joint_positions: Mapping[str, float] = None,
        ignore_joints_with_str: List[str] = None,
    ):
        """Visualizer for a PybulletSimulatorInterface instance to test joint positioning etc.

        When using with a PybulletRobot or child class object, use the class method `fromBulletRobot`
        to make things easier.

        See `examples/utils_demo/demo_bullet_robot_visualizer.py`

        WARNING
        -------
        DO NOT use other GUI utils for pybullet (other sliders, buttons etc), when
        using this. This class uses the `removeAllUserParameters` parameters exposed
        in pybullet, which means it can remove every debug item created externally.

        TODO: fix this to only remove self-created items.

        Args:
            pb_sim_interface (PybulletSimulatorInterface): The simulator interface instance.
            joint_lower_lims (Mapping[str, float]): Mapping from joint names to joint lower limits.
            joint_upper_lims (Mapping[str, float]): Mapping from joint names to joint upper limits.
            starting_joint_positions (Mapping[str, float], optional): Mapping from joint names to
                starting positions to be used during the visualisation. Defaults to zeros.
            ignore_joints_with_str (List[str], optional): List of substrings to look for in joint
                names, matching joints will be ignored from the joint sliders GUI (useful for
                joints such as wheels). Defaults to ["_wheel"] if None provided.
        """
        self.sim_robot = pb_sim_interface
        self._init_j = starting_joint_positions if starting_joint_positions is not None else {}
        self._ignore_joints_with_str = (
            ["_wheel"] if ignore_joints_with_str is None else ignore_joints_with_str
        )
        self._default_base_pose = copy.deepcopy(self.sim_robot.get_base_pose())
        self._base_pose_target = self.sim_robot.get_base_pose()

        def names_match(name, list_of_words):
            for word in list_of_words:
                if word in name:
                    return True
            return False

        self._enable_stepping = False
        self._j_names = []
        self._j_lower_lims = []
        self._j_upper_lims = []
        for jname in joint_lower_lims.keys():
            if names_match(jname, self._ignore_joints_with_str):
                continue
            self._j_names.append(jname)
            self._j_lower_lims.append(joint_lower_lims[jname])
            self._j_upper_lims.append(joint_upper_lims[jname])
        self._ok_to_run = True
        self._read_button_thread = None

    def run(
        self,
        sim_step_rate: float = 240,
        slider_update_rate: float = 10,
        additional_callbacks: List[Callable[[], None]] = None,
    ):
        """Run the visualizer thread (non-blocking).

        Args:
            sim_step_rate (float, optional): The desired rate (Hz) for calling step simulation
                (This is not the timestep used in the physics of the simulation). Defaults to 240
                (technically this should make it realtime).
            slider_update_rate (float, optional): Update rate (Hz) for reading the joint sliders in
                pybullet GUI to update robot states. Defaults to 10.
        """

        gui_cbs: Dict[str, PbGUICallback] = {}

        if additional_callbacks is None:
            additional_callbacks = []

        for cb in additional_callbacks:
            assert callable(cb)

        if self.sim_robot._in_torque_mode:
            logging.warning(
                "PybulletRobot should be in position control mode to use this visualiser."
                " Changing to position control mode now."
            )
            self.sim_robot.set_position_control_mode()

        self._prev_jvals: np.ndarray = None

        # callbacks for buttons and sliders in the pybullet gui
        def _set_j_cb(j_vals: List[float]):
            if self._prev_jvals is not None and np.allclose(j_vals, self._prev_jvals):
                pass
            else:
                self.sim_robot.reset_actuated_joint_positions(j_vals, self._j_names)
                # setting position control command as well (needed if simulation is switched on)
                self.sim_robot.set_actuated_joint_commands(
                    q=j_vals, actuated_joint_names=self._j_names
                )
                self._prev_jvals = copy.deepcopy(j_vals)

        def _set_base_target_cb(pose_target: List[float]):
            # live update of base pose not being done to allow sim enabling
            self._base_pose_target = (
                pose_target[:3],
                rpy2quat(np.deg2rad(pose_target[3:])),
            )

        def _set_base_pose_cb():
            self.sim_robot.reset_base_pose(self._base_pose_target[0], self._base_pose_target[1])

        def _reset_base_cb():
            self.sim_robot.reset_base_pose(self._default_base_pose[0], self._default_base_pose[1])

        def _place_on_ground_cb():
            jvals = self.sim_robot.get_actuated_joint_positions()
            move_resolution = 0.01
            if not np.any(self.sim_robot.get_contact_states_of_links(self.sim_robot.link_ids)):
                bpos = self.sim_robot.get_base_pose()[0]
                in_collision = False
                while not in_collision:
                    bpos[2] -= move_resolution
                    self.sim_robot.reset_base_pose(
                        position=bpos,
                        orientation=self.sim_robot.get_base_pose()[1],
                    )
                    self.sim_robot.step()
                    in_collision = np.any(
                        self.sim_robot.get_contact_states_of_links(self.sim_robot.link_ids)
                    )

            self.sim_robot._place_robot_on_ground(
                move_resolution=move_resolution,
                default_position=self.sim_robot.get_base_pose()[0],
                default_orientation=self.sim_robot.get_base_pose()[1],
            )
            self.sim_robot.reset_actuated_joint_positions(jvals)

        def _toggle_sim():
            self._enable_stepping = not self._enable_stepping
            if self._enable_stepping:
                # otherwise joints positions will drift
                self.sim_robot.set_actuated_joint_commands(
                    q=self.sim_robot.get_actuated_joint_positions(
                        self.sim_robot.actuated_joint_names
                    ),
                    actuated_joint_names=self.sim_robot.actuated_joint_names,
                )

        def _reset_debug_params(j_mode: str = "default"):
            """Fully reset all GUI stuff (sliders, buttons)"""

            pb.removeAllUserParameters(self.sim_robot.cid)
            ## Below line does not work. pb.removeUserDebugItem does not seem to
            ## work for buttons and sliders!
            # for gui_name in gui_cbs.keys():
            #     gui_cbs[gui_name].remove_gui_object()

            gui_cbs["reset_joints_button"] = PbGUIButtonCallback(
                button_name="reset joints",
                cid=self.sim_robot.cid,
                callback=partial(_reset_debug_params, j_mode="default"),
            )
            gui_cbs["random_joints_button"] = PbGUIButtonCallback(
                button_name="randomise joints",
                cid=self.sim_robot.cid,
                callback=partial(_reset_debug_params, j_mode="random"),
            )
            gui_cbs["zero_joints_button"] = PbGUIButtonCallback(
                button_name="zero all joints",
                cid=self.sim_robot.cid,
                callback=partial(_reset_debug_params, j_mode="zero"),
            )
            j_default_vals = []
            for n, j_name in enumerate(self._j_names):
                if j_mode == "zero":
                    j_default_vals.append(
                        0
                        if self._j_lower_lims[n] <= 0.0 <= self._j_upper_lims[n]
                        else self._j_lower_lims[n]
                    )
                elif j_mode == "random":
                    j_default_vals.append(
                        np.random.uniform(self._j_lower_lims[n], self._j_upper_lims[n])
                    )
                else:
                    j_default_vals.append(0 if j_name not in self._init_j else self._init_j[j_name])
            gui_cbs["triggered_joint_sliders"] = RateTriggeredMultiCallbacks(
                gui_callbacks=[
                    PbMultiGUISliderSingleCallback(
                        slider_names=self._j_names,
                        slider_lower_lims=self._j_lower_lims,
                        slider_upper_lims=self._j_upper_lims,
                        slider_default_vals=j_default_vals,
                        callback=_set_j_cb,
                        cid=self.sim_robot.cid,
                    )
                ],
                rate=slider_update_rate,
            )
            gui_cbs["place_on_ground_button"] = PbGUIButtonCallback(
                button_name="place robot on ground",
                cid=self.sim_robot.cid,
                callback=_place_on_ground_cb,
            )
            gui_cbs["run_sim_button"] = PbGUIButtonCallback(
                button_name="enable/disable simulation",
                cid=self.sim_robot.cid,
                callback=_toggle_sim,
            )

            starting_pos, starting_ori = self.sim_robot.get_base_pose()
            starting_rpy = np.rad2deg(quat2rpy(starting_ori))
            gui_cbs["set_base_sliders"] = PbMultiGUISliderSingleCallback(
                slider_names=[
                    "base_x_pos",
                    "base_y_pos",
                    "base_z_pos",
                    "base_roll",
                    "base_pitch",
                    "base_yaw",
                ],
                slider_lower_lims=[-1, -1, -1, -180, -180, -180],
                slider_upper_lims=[1, 1, 1, 180, 180, 180],
                slider_default_vals=starting_pos.tolist() + starting_rpy.tolist(),
                callback=_set_base_target_cb,
                cid=self.sim_robot.cid,
            )
            gui_cbs["set_base_button"] = PbGUIButtonCallback(
                button_name="set base pose",
                cid=self.sim_robot.cid,
                callback=_set_base_pose_cb,
            )
            gui_cbs["reset_base_button"] = PbGUIButtonCallback(
                button_name="reset base",
                cid=self.sim_robot.cid,
                callback=_reset_base_cb,
            )

        def _read_params_loop():
            """The main gui update loop thread"""

            _reset_debug_params()
            rate_limiter = RateLimiter(sim_step_rate, warn=False)
            logging.info(f"Starting {self.__class__.__name__} instance...")
            while self._ok_to_run:
                for cb in gui_cbs.values():
                    cb.run_once()

                if self._enable_stepping:
                    self.sim_robot.step()

                for cb in additional_callbacks:
                    cb()

                rate_limiter.sleep()

        self._read_button_thread = threading.Thread(target=_read_params_loop)
        self._read_button_thread.start()
        logging.info(f"{self.__class__.__name__} running...")

    def close(self):
        """Close the visualizer."""
        self._ok_to_run = False
        logging.info(f"Attempting to shut down {self.__class__.__name__} instance.")
        if self._read_button_thread is not None:
            self._read_button_thread.join()
        logging.info(f"Closed {self.__class__.__name__} instance.")

    @classmethod
    def fromBulletRobot(
        cls: "PybulletRobotVisualizer",
        pb_robot: PybulletRobot,
        starting_joint_positions: Mapping[str, float] = None,
        ignore_joints_with_str: List[str] = None,
    ) -> "PybulletRobotVisualizer":
        """Visualizer for a PybulletRobot instance to test joint positioning etc.

        See `examples/utils_demo/demo_bullet_robot_visualizer.py`

        Args:
            pb_robot (PybulletRobot): The PybulletRobot instance to load.
            starting_joint_positions (Mapping[str, float], optional): Mapping from joint names to
                starting positions to be used during the visualisation. Defaults to zeros.
            ignore_joints_with_str (List[str], optional): List of substrings to look for in joint
                names, matching joints will be ignored from the joint sliders GUI (useful for
                joints such as wheels). Defaults to ["_wheel"] if None provided.
        """
        pin = pb_robot.get_pinocchio_interface()
        return cls(
            pb_sim_interface=pb_robot._sim_robot,
            joint_lower_lims=dict(zip(pin.actuated_joint_names, pin.actuated_joint_lower_limits)),
            joint_upper_lims=dict(zip(pin.actuated_joint_names, pin.actuated_joint_upper_limits)),
            starting_joint_positions=(
                starting_joint_positions
                if (starting_joint_positions is not None and starting_joint_positions != {})
                else dict(
                    zip(
                        pb_robot._sim_robot.actuated_joint_names,
                        pb_robot._sim_robot.get_actuated_joint_positions(),
                    )
                )
            ),
            ignore_joints_with_str=ignore_joints_with_str,
        )



"""
Base class for creating user interfaces (UI) to act as fake Global Plan generators.

This interface is for generating a fake global plan to be used by local planners from user
input and 'translating' the input to references that can be used by local planners for generating
control references.

In fully autonomous systems, this will be replaced by a true global planner.

e.g. read joystick input and translate to base velocity commands (as GlobalMotionPlan message)
"""

from abc import abstractmethod

from ..global_planner_base import GlobalMotionPlannerBase
from ....core.types import GlobalMotionPlan, RobotState


class UIBase(GlobalMotionPlannerBase):

    @abstractmethod
    def process_user_input(
        self, robot_state: RobotState, t: float = None, dt: float = None
    ) -> GlobalMotionPlan:
        """Processes user input and returns a GlobalMotionPlan object.

        The method that should be implemented by child classes where a user input is processed
        and appropriate GlobalMotionPlan is returned.

        Returns:
            GlobalMotionPlan: the generated motion plan object.
        """
        raise NotImplementedError("This method has to be implemented in the child class")

    def generate_global_plan(
        self, robot_state: RobotState, t: float = None, dt: float = None
    ) -> GlobalMotionPlan:
        """This method is following the global planner protocol.

        Internally this just calls the `process_user_input()` method.

        Returns:
            GlobalMotionPlan: the generated motion plan object.
        """
        return self.process_user_input(robot_state=robot_state, t=t, dt=dt)


class DummyUI(UIBase):
    def __init__(self, squawk: bool = True):
        """Dummy UI/Global planner class for testing pipeline.

        Args:
            squawk (bool, optional): Verbose. Defaults to True.
        """
        self._squawk = squawk

    def process_user_input(
        self, robot_state: RobotState, t: float = None, dt: float = None
    ) -> GlobalMotionPlan:
        if self._squawk:
            print(f"{__class__.__name__}: Faking user input read step.")
        return GlobalMotionPlan()



"""Defines controller managers that can be used to handle multiple controllers
in the control loop."""

from .controller_manager_base import ControllerManagerBase
from .simple_controller_manager import SimpleControllerManager



from abc import abstractmethod
import numpy as np

from ..robot_interface_with_pinocchio import RobotInterfaceWithPinocchio
from ....utils.time_utils import ClockBase
from ....core.types import QuatType


class SimulatedRobotInterface(RobotInterfaceWithPinocchio):
    """Base class for simulated robots.

    All simulated robot classes should also expose a the following methods
    in addition to the standard read() and write() methods:

    1. set_base_pose: This is so that we can use proxy base controlling
        (for mobile manipulators in sim that need stable base motion
        (which may not be ready for testing)).
    2. get_sim_clock: this is to expose the time in the simulated world.
        This can then be used so that all components in the control
        loop can use the same clock from the sim world, and makes the
        velocity measurements in the sim interface more sensible wrt
        the clock being used in the control loop. (see example 05 to
        see how this clock is used in the control loop.)
    """

    @abstractmethod
    def set_base_pose(self, position: np.ndarray, orientation: QuatType):
        """
        Set the pose of the base of the simulated robot in the world frame.

        Args:
            position (np.ndarray): Desired position for the base frame in the world.
            orientation (QuatType): Desired orientation quaternion [x,y,z,w] for base frame
                in the world.

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.
        """
        raise NotImplementedError("This method has to be implemented in the child class")

    @abstractmethod
    def get_sim_clock(self) -> ClockBase:
        """
        Get the Clock object that gives the time in this simulated world.

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.

        Returns:
            ClockBase: The Clock object that can be used to get the current time using its
                get_time() method.
        """
        raise NotImplementedError("This method has to be implemented in the child class")



"""Base class defining the interface for all state estimators to be used in the control loop."""

from abc import abstractmethod

from ..pyrcf_component import PyRCFComponent
from ...core.types import RobotState


class StateEstimatorBase(PyRCFComponent):  # pylint: disable=too-few-public-methods
    """Base class defining the interface for all state estimators to be used in the control loop."""

    @abstractmethod
    def update_robot_state_with_state_estimates(
        self,
        robot_state: RobotState,
        t: float = None,
        dt: float = None,
    ) -> RobotState:
        """The main state estimator update method to be called in the loop.

        This method will be called after the proprioceptive data in RobotState is retreived by
        calling the `RobotInterface->read()` method.

        This method should be implemented by any state estimator implementation.

        Args:
            robot_state (RobotState): The current state information from the robot.
            t (float, optional): the current time signature of the control loop. Defaults to None
                (controllers may or may not need this).
            dt (float, optional): the time since the last control loop. Defaults to None
                (controllers may or may not need this).

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.

        Returns:
            RobotState: Return the same robot state object with the fields for
                robot_state.state_estimates filled appropriately.
        """
        raise NotImplementedError("This method has to be implemented in the child class")


class DummyStateEstimator(StateEstimatorBase):  # pylint: disable=too-few-public-methods
    """Dummy state estimator class for testing pipeline."""

    def __init__(self, squawk: bool = True):
        """Dummy state estimator class for testing pipeline.

        Args:
            squawk (bool, optional): Verbose. Defaults to True.
        """
        self._squawk = squawk

    def update_robot_state_with_state_estimates(
        self,
        robot_state: RobotState,
        t: float = None,
        dt: float = None,
    ) -> RobotState:
        """This is a dummy state estimator that does nothing at all and returns the original state
        object back as is."""
        if self._squawk:
            print(f"{__class__.__name__}: Estimating (fake) robot states.")
        return robot_state



"""Defines agents that can be used to compute control commands to be sent to
a robot given a global plan (target/task). When using 'classical' methods,
an agent may just be a local planner + a controller (use the PlannerControllerAgent in
this case). When using machine-learned controller, this could be a policy that direcly
uses the global plan and outputs a robot command (implement an agent from MLAgentBase)."""

from .agent_base import AgentBase, DummyAgent
from .planner_controller_agent import PlannerControllerAgent
from .ml_agent_base import MLAgentBase
from .torchscript_agent_base import TorchScriptAgentBase



from abc import abstractmethod
from typing import Tuple, Literal
import numpy as np

from ...core.types import GlobalMotionPlan, RobotCmd, RobotState, JointStates, LocalMotionPlan
from .ml_agent_base import MLAgentBase
from ...core.logging import logging
from ...utils.time_utils import ClockBase, PythonPerfClock, RateTrigger

from ...variables import TORCH_AVAILABLE

if TORCH_AVAILABLE:
    import torch  # pylint: disable=E0401


class TorchScriptAgentBase(MLAgentBase):
    """A (abstract) base class that provides functionalities to use a torchscript model file
    and use it's inference mode to perform control update.

    Child class has to override two methods:
        - update_input_to_model
        - update_cmd_from_model_output
    See docstrings for each method in MLAgentBase class.
    """

    def __init__(
        self,
        model_file: str,
        input_dims: int,
        device: Literal["cpu", 0] = "cpu",
        warmup_iterations: int = 10,
        default_kp: np.ndarray | float = 20,
        default_kd: np.ndarray | float = 1.0,
        update_rate: float = None,
        clock: ClockBase = PythonPerfClock(),
        dtype: "torch.dtype" = None,
    ):
        """A (abstract) base class that provides functionalities to use a torchscript model file
        and use it's inference mode to perform control update.

        Child class has to override two methods:
            - update_input_to_model
            - update_cmd_from_model_output
        See docstrings for each method.

        Args:
            model_file (str): Path to torchscript (jit) file
            input_dims (int): Dimensions of the input vector to the model.
            device (Literal["cpu", 0] = "cpu", optional): Torch device to load tensors and model onto.
                Defaults to "cpu".
            warmup_iterations (int, optional): Number of iterations to send dummy input to the model
                on construction (for warming up model/processors). Defaults to 10.
            default_kp (np.ndarray | float, optional): Default joint positions gains. Defaults to 20.
            default_kd (np.ndarray | float, optional): Default joint damping gains. Defaults to 1.0.
            update_rate (float, optional): Update rate for the RL controller (if required to run at
                lower rate than the control loop). Use factors of the control loop's frequency.
                Defaults to None (run at the control loop rate).
            clock (ClockBase, optional): The clock to use for timer. Defaults to PythonPerfClock().
            dtype (torch.dtype, optional): Datatype of model parameters and tensors. Defaults to
                `torch.float`.
        """
        if not TORCH_AVAILABLE:
            raise RuntimeError(
                f"Pytorch has to be installed for the {self.__class__.__name__} class to be used."
                " e.g. `pip install torch`."
            )

        if dtype is None:
            dtype = torch.float

        self._torch_device = torch.device(device)
        model = torch.jit.load(model_file, map_location=self._torch_device)
        model.eval()
        self._model = torch.jit.optimize_for_inference(model)

        self._device = device
        self._dtype = dtype
        self._dims = input_dims
        self._tensor_inp = torch.tensor(
            np.zeros(self._dims), dtype=self._dtype, device=self._torch_device
        ).reshape([1, -1])
        self._input_ndarray = np.zeros(self._dims)
        """The input array to be updated by the child class using the current robot
        state and global plan commands. This will be passed to the model as input during
        inference"""

        if warmup_iterations is not None and warmup_iterations > 0:
            logging.info("Warming up policy...")
            for _ in range(warmup_iterations):
                self._model.forward(self._tensor_inp)
            logging.info("Warm up complete.")

        self._latest_ctrl_cmd: RobotCmd = None
        """The RobotCmd object that can be use in the `update_input_to_model` method
        (if required) as the last sent command to the robot. This object has to be updated
        by the child class's `update_cmd_from_model_output` using the output from the model
        inference step."""

        self._default_kp = default_kp
        self._default_kd = default_kd

        self._rate = update_rate
        if self._rate is not None and self._rate > 0.0:
            self._rate_trigger = RateTrigger(rate=self._rate, clock=clock)
            logging.debug(f"{self.__class__.__name__}: Setting trigger rate to {self._rate}Hz.")

    def _should_run(self):
        if self._rate is None:
            return True
        if self._rate <= 0.0:
            return False
        return self._rate_trigger.triggered()

    def initialise_robot_cmd(self, joint_states: JointStates):
        self._latest_ctrl_cmd: RobotCmd = RobotCmd.fromJointStates(
            joint_states=joint_states,
            Kp=self._default_kp,
            Kd=self._default_kd,
        )
        self._latest_ctrl_cmd.joint_commands.joint_efforts *= 0.0
        self._latest_ctrl_cmd.joint_commands.joint_velocities *= 0.0

    @abstractmethod
    def update_input_to_model(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> np.ndarray:
        """Should update (`self._input_ndarray`) using appropriate values (input to model). This method
        has access to `self._latest_ctrl_cmd` (type `RobotCmd`) as well if needed.
        (NOTE: `self._latest_ctrl_cmd` is set to be the initial robot joint positions (zero velocities
        and efforts commands) with `default_kp` and `default_kd` at start).
        """

    @abstractmethod
    def update_cmd_from_model_output(
        self,
        model_output: np.ndarray,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> None:
        """Should update `self._latest_robot_cmd` (type `RobotCmd`) using the output from the NN model.

        Args:
            model_output (np.ndarray): the numpy array created from the output tensor from the model
                after the inference query was done. This is the output of the neural network. This
                method should use this object to update `self._latest_robot_cmd` to be sent to the
                robot.
        """

    def get_action(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float = None,
        dt: float = None,
    ) -> RobotCmd:
        if self._latest_ctrl_cmd is None:
            self.initialise_robot_cmd(joint_states=robot_state.joint_states)
        # only perform inference and command update if required rate is met
        if self._should_run():
            # update the input tensor to the model by calling the child class's
            # `update_input_to_model` method.
            self.update_input_to_model(robot_state=robot_state, global_plan=global_plan, t=t, dt=dt)
            # update the pre-loaded tensor with the values from the updated
            # `self._input_ndarray` object.
            for idx in range(self._dims):
                self._tensor_inp[0, idx] = self._input_ndarray[idx]

            # perform inference using this input tensor
            with torch.inference_mode():
                output_tensor = self._model.forward(self._tensor_inp)

            # use the child class's `update_cmd_from_model_output` method
            # to update the RobotCmd object that is to be returned to the
            # control loop for writing to the robot
            self.update_cmd_from_model_output(
                model_output=output_tensor.cpu().detach().numpy().flatten(),
                robot_state=robot_state,
                global_plan=global_plan,
                t=t,
                dt=dt,
            )

        return self._latest_ctrl_cmd

    def get_last_output(self) -> Tuple[LocalMotionPlan, RobotCmd]:
        return None, self._latest_ctrl_cmd



## CALLBACK FUNCTIONALITY UTILS WITH PbGUIs
from typing import Tuple, List, Callable, TypeAlias
import numpy as np

from .base_callbacks import CustomCallbackBase
from .pb_gui_utils import (
    PybulletGUISlider,
    PybulletGUIButton,
    PybulletDebugFrameViz,
    PybulletDebugPoints,
)

QuatType: TypeAlias = np.ndarray
"""Numpy array representating quaternion in format [x,y,z,w]"""
Vector3D: TypeAlias = np.ndarray
"""Numpy array representating 3D cartesian vector in format [x,y,z]"""


class PbGUICallback(CustomCallbackBase):
    """Abstraction for all GUICallbacks defined using Pybullet GUI."""


class PbGUISliderCallback(PbGUICallback):
    """Create a slider in pybullet and specify the callback to apply the slider value to."""

    def __init__(
        self,
        slider_name: str,
        slider_lower_lim: float,
        slider_upper_lim: float,
        slider_default_val: float,
        callback: Callable[[float], None],
        cid: int,
        run_only_if_value_changed: bool = True,
        mapping_function: Callable[[float], float] = lambda x: x,
    ):
        """Create a slider in pybullet and specify the callback to apply the slider value to.

        Args:
            slider_name (str): Name for slider.
            slider_lower_lim (float): The min value for slider.
            slider_upper_lim (float): The max value for slider.
            slider_default_val (float): Default starting value for slider.
            callback (Callable[[float], None]): Handle to a function that takes as input
                the slider value (no other arguments).
            cid (int): The client id to connect to the physics simulator.
            run_only_if_value_changed (bool, optional): The callback will be executed only if
                slider value has changed since the last time it was checked (i.e. `run_once`
                was called for this object).
            mapping_function (Callable[[float], float], optional): Mapping to do for the
                value read from the slider (e.g. convert radian to degrees)
        """
        self._slider = PybulletGUISlider(
            name=slider_name,
            lower_lim=slider_lower_lim,
            upper_lim=slider_upper_lim,
            default_val=slider_default_val,
            cid=cid,
            mapping_function=mapping_function,
        )
        self._check_first = run_only_if_value_changed
        self._callback = callback

    def run_once(self) -> None:
        """This method has to be called for the callback to be executed."""
        if self._check_first and not self._slider.value_changed():
            return
        self._callback(self._slider.get_value())

    def cleanup(self):
        self._slider.remove()


class PbMultiGUISliderSingleCallback(PbGUICallback):
    """Create multiple sliders in pybullet and use values from all the sliders in a single
    callback."""

    def __init__(
        self,
        slider_names: List[str],
        slider_lower_lims: List[float],
        slider_upper_lims: List[float],
        slider_default_vals: List[float],
        callback: Callable[[List[float]], None],
        cid: int,
        run_only_if_value_changed: bool = True,
        mapping_functions: Callable[[float], float] | List[Callable[[float], float]] = lambda x: x,
    ):
        """Create multiple sliders and use values from all the sliders in a single callback.

        Args:
            slider_names (List[str]): Names for slider.
            slider_lower_lims (List[float]): Corresponding min values for sliders.
            slider_upper_lims (List[float]): Corresponding max values for sliders.
            slider_default_vals (List[float]): Corresponding default values for sliders.
            callback (Callable[[List[float]], None]): Handle to a function that takes as input
                the slider values (list of floats).
            cid (int): The client id to connect to the physics simulator.
            run_only_if_value_changed (bool, optional): The callback will be executed only if
                slider value has changed since the last time it was checked (i.e. `run_once`
                was called for this object).
            mapping_functions (Callable[[float], float] | List[Callable[[float], float]], optional):
                Mapping to do for the value read from the slider (e.g. convert radian to degrees).
        """
        self._sliders: List[PybulletGUISlider] = []
        if callable(mapping_functions):
            mapping_functions = [mapping_functions] * len(slider_names)
        for n, name in enumerate(slider_names):
            self._sliders.append(
                PybulletGUISlider(
                    name=name,
                    lower_lim=slider_lower_lims[n],
                    upper_lim=slider_upper_lims[n],
                    default_val=slider_default_vals[n],
                    cid=cid,
                    mapping_function=mapping_functions[n],
                )
            )
        self._check_first = run_only_if_value_changed
        self._callback = callback

    def run_once(self):
        """This method has to be called for the callback to be executed."""
        if self._check_first:
            for s in self._sliders:
                if s.value_changed():
                    break
            else:
                return
        vals = [slider.get_value() for slider in self._sliders]
        self._callback(vals)

    def cleanup(self):
        for slider in self._sliders:
            slider.remove()
        self._sliders.clear()

    def get_sliders(self) -> List[PybulletGUISlider]:
        """Get a list of all slider objects.

        Returns:
            List[PybulletGUISlider]: List of slider objects.
        """
        return self._sliders


class PbGUIButtonCallback(PbGUICallback):
    """Create a button in pybullet and specify the callback to apply the slider value to."""

    def __init__(self, button_name: str, callback: Callable[[], None], cid: int):
        """Create a button in pybullet and specify the callback to apply the slider value to.

        Args:
            button_name (str): Name of button.
            callback (Callable[[], None]): Handle to a function that should be executed when
                the button is pressed.
            cid (int): The client id to connect to the physics simulator.
        """
        self._button = PybulletGUIButton(name=button_name, cid=cid)
        self._callback = callback

    def run_once(self):
        """This method has to be called for checking button press and for the
        callback to be executed."""
        if self._button.was_pressed():
            self._callback()

    def cleanup(self):
        self._button.remove()


class PbDebugFrameVizCallback(PbGUICallback):
    """Draw a coordinate frame in pybullet that changes its pose based on the pose
    returned by a callback function."""

    def __init__(
        self,
        callback: Callable[[], Tuple[Vector3D, QuatType]],
        cid: int,
        line_length: float = 0.2,
        line_width: float = 2,
        duration: float = 0.0,
    ):
        """Draw a coordinate frame in pybullet that changes its pose based on the pose
        returned by a callback function.

        Args:
            callback (Callable[[Vector3D, QuatType], None]): Handle to a function that returns
                a position and orientation target for the coordinate frame to visualize.
            cid (int): The physics server id to connect to.
            line_length (float, optional): Length of the lines in the world. Defaults to 0.2.
            line_width (float, optional): Line thickness. Defaults to 2.
            duration (float, optional): Lifetime for the frame. Defaults to 0.0 (infinite).
        """
        self._frame_viz = PybulletDebugFrameViz(
            cid=cid,
            position=None,
            orientation=None,
            line_width=line_width,
            line_length=line_length,
            duration=duration,
            draw_at_init=False,
        )
        assert callable(callback)
        self._callback = callback

    def run_once(self):
        """This method has to be called for for getting the pose from the callback
        function and updating the frame visualiser's pose."""
        p, q = self._callback()
        self._frame_viz.update_frame_pose(position=p, orientation=q)

    def cleanup(self):
        self._frame_viz.remove()


class PbDebugPointsCallback(PbGUICallback):
    """Visualise point(s)/sphere(s) in pybullet exposing a callback to change its position
    in the world.
    """

    def __init__(
        self,
        callback: Callable[[], List[Vector3D]],
        cid: int,
        point_positions: List[Vector3D] = None,
        point_size: float = 1.0,
        rgb: Tuple[float, float, float] = (1, 0, 0),
        lifetime: float = 0.0,
    ):
        """Visualise point(s)/sphere(s) in pybullet exposing a callback to change its position
        in the world.

        Args:
            callback (Callable[[], List[Vector3D]]): Function returning positions of the point.
            cid (int): The physics server id to connect to.
            point_positions (List[Vector3D], optional): Starting positions for the points. Defaults
                to None.
            point_size (float, optional): Size of points. Defaults to 1.0.
            rgb (Tuple[float, float, float], optional): RGB color code. Defaults to (1, 0, 0).
            lifetime (float, optional): Lifetime for the points in seconds. Defaults to 0.0
                (persistent).
        """
        self._points_viz = PybulletDebugPoints(
            point_positions=point_positions,
            point_size=point_size,
            rgb=rgb,
            lifetime=lifetime,
            cid=cid,
        )
        assert callable(callback)
        self._callback = callback

    def run_once(self):
        self._points_viz.update_point_positions(point_positions=self._callback())

    def cleanup(self):
        self._points_viz.remove()



"""A simple keyboard interface following the GlobalMotionPlanner protocol."""

from typing import Callable, Dict, Tuple
import copy
import threading
import pygame

from .ui_base import UIBase
from ....core.types import GlobalMotionPlan, RobotState
from .key_mappings import DEFAULT_KEYBOARD_MAPPING
from .ui_utils import get_keymapping_doc


def blit_multiline_text(surface, text, pos, font, line_spacing=3, color=pygame.Color("white")):
    """Function to display multiline text in pygame screen."""
    # 2D array where each row is a list of words.
    words = [word.split(" ") for word in text.splitlines()]
    x, y = pos
    for line in words:
        for word in line:
            tab_words = word.split("\t")
            for twd in tab_words:
                if twd == "":
                    twrd = "    "
                else:
                    twrd = twd
                word_surface = font.render(twrd, 0, color)
                word_width, word_height = word_surface.get_size()
                if x + word_width >= surface.get_size()[0]:
                    x = pos[0]  # Reset the x
                    y += word_height  # Start on new row
                surface.blit(word_surface, (x, y))
                x += word_width + font.size(" ")[0]
        x = pos[0]  # Reset the x
        y += word_height + line_spacing  # Start on new row


class KeyboardInterface(UIBase):
    """A simple keyboard interface following the GlobalMotionPlanner protocol."""

    def __init__(
        self,
        key_mappings: Dict[str, Callable[[GlobalMotionPlan], GlobalMotionPlan]] = None,
        default_global_plan: GlobalMotionPlan = GlobalMotionPlan(),
        window_size: Tuple[int, int] = (1000, 1000),
        parallel_mode: bool = False,
        verbose: bool = False,
    ) -> None:
        """A simple keyboard interface following the GlobalMotionPlanner protocol.

        This class allows using keybindings to update a global motion plan from user keyboard input.
        The modified global motion plan object can then be retrieved when generate_global_plan
        method is called in the control loop.

        NOTE: The pygame window should be in focus for the inputs to be recorded by this class.
        NOTE: This component initiates a thread for reading keyboard input!

        Args:
            key_mappings (dict[ str, callable[[GlobalMotionPlan], GlobalMotionPlan] ], optional):
                A char --> callable dictionary mapping a key from the keyboard to a function that
                updates a GlobalMotionPlan object. Defaults to DEFAULT_KEYBOARD_MAPPING.
            default_global_plan (GlobalMotionPlan, optional): the default to use when initialising
                (and when reset it called (to be implemented)). Defaults to GlobalMotionPlan().
            window_size (Tuple[int, int]): pygame window size in pixels. Defaults to (500, 500).
            parallel_mode (bool): If set to True, this will run in a separate thread (continuously
                recording and storing last user input).
            verbose (bool): If true, will print command on console every time user inputs a valid
                key.
        """
        self._global_plan = default_global_plan
        self._key_mappings = key_mappings if key_mappings is not None else DEFAULT_KEYBOARD_MAPPING
        self._parallel_mode = parallel_mode
        self._verbose = verbose

        pygame.init()
        pygame.mixer.quit()  # disable audio (otherwise alsa overrun warnings)

        self._display = pygame.display.set_mode(window_size)
        pygame.display.set_caption("Keyboard input -- mock global planner")

        self._font = pygame.font.SysFont("Ubuntu", 12)

        self._doc_text, self._key_docs = get_keymapping_doc(self._key_mappings)

        blit_multiline_text(self._display, self._doc_text, (10, 10), self._font)

        pygame.display.update()

        if self._parallel_mode:
            self._keyboard_thread = threading.Thread(
                target=self._update_global_plan_from_user_input
            )
            self._keyboard_thread.start()

    def _update_global_plan_from_user_input(self) -> None:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()

            if event.type == pygame.KEYDOWN:
                c = pygame.key.name(event.key)
                if c in self._key_mappings:
                    self._global_plan = self._key_mappings[c](self._global_plan)
                    msg = f"{self.__class__.__name__}: User input received: Key {c}\n"
                    msg += f"{self._key_docs[c]}\n\nNew global plan: {self._global_plan}"
                    if self._verbose:
                        print(msg)
                    self._display.fill((0, 0, 0))
                    blit_multiline_text(
                        self._display,
                        self._doc_text + "\n\n" + msg,
                        (10, 10),
                        self._font,
                    )
                    pygame.display.update()

    def process_user_input(
        self, robot_state: RobotState, t: float = None, dt: float = None
    ) -> GlobalMotionPlan:
        if self._global_plan.joint_references.joint_names is None:
            self._global_plan.joint_references.joint_names = copy.deepcopy(
                robot_state.joint_states.joint_names
            )
            self._global_plan.joint_references.joint_positions = copy.deepcopy(
                robot_state.joint_states.joint_positions
            )
        if not self._parallel_mode:
            self._update_global_plan_from_user_input()
        return self._global_plan

    def shutdown(self):
        super().shutdown()
        try:
            pygame.quit()
        except TypeError:
            pass
        if self._parallel_mode:
            self._keyboard_thread.join()



from abc import abstractmethod
from typing import Tuple

from ..pyrcf_component import PyRCFComponent
from ...core.types import RobotState, RobotCmd, GlobalMotionPlan, LocalMotionPlan


class AgentBase(PyRCFComponent):
    """An agent is an entity that observes the state of the robot and responds with a control
    action.

    With this in mind, an agent for classical control cases would be an entity that combines the
    functions of a 'local planner' and 'controller' into a single entity called 'Agent', which
    observes the state of the robot interface and responds by sending commands to the robot, so as
    to follow the objective of following the 'global plan' from the global planner.
    """

    @abstractmethod
    def get_action(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> RobotCmd:
        """Compute the control command to be executed given the current state and the global plan.

        Args:
            robot_state (RobotState): Current robot state. This is equivalent to 'observation' in
                standard learning-based agents.
            global_plan (GlobalMotionPlan): The reference global plan to follow.
            t (float, optional): the current time signature of the control loop. Defaults to None
                (controllers may or may not need this).
            dt (float, optional): the time since the last control loop. Defaults to None
                (controllers may or may not need this).

        Returns:
            RobotCmd: The output control command (action) to be sent to the robot.

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.
        """
        raise NotImplementedError("This method has to be implemented in the child class")

    @abstractmethod
    def get_last_output(self) -> Tuple[LocalMotionPlan, RobotCmd]:
        """Should return the last computed outputs by this agent.

        Returns:
            Tuple[LocalMotionPlan, RobotCmd]: Output local plan and control command from this agent.
                NOTE: Either of these can be None, if the agent does not compute them.
        """
        raise NotImplementedError("This method has to be implemented in the child class")


class DummyAgent(AgentBase):
    """Dummy Agent for testing pipeline."""

    def __init__(self, squawk: bool = True):
        """Dummy Agent for testing pipeline.

        Args:
            squawk (bool, optional): Verbosity. Defaults to True.
        """
        self._squawk = squawk

    def get_action(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> RobotCmd:
        """This is a dummy method for sanity checking the control loop."""
        if self._squawk:
            print(f"{__class__.__name__}: Performing dummy agent update. time: {t}, dt: {dt}")
        return RobotCmd()

    def get_last_output(self) -> Tuple[LocalMotionPlan | RobotCmd]:
        return None, RobotCmd()



import importlib
import importlib.util


def package_available(package_name: str):
    return importlib.util.find_spec(package_name) is not None


# optional third party libraries
TORCH_AVAILABLE = package_available("torch")


__all__ = ["TORCH_AVAILABLE"]



from abc import ABC, abstractmethod
from typing import List

from ...core.types import GlobalMotionPlan, RobotCmd, RobotState
from ..agents.agent_base import AgentBase
from .command_accumulators.cmd_accumulation_policies import CommandAccumulatorBase


class ControllerManagerBase(ABC):
    """Base class for writing controller (agent) managers."""

    def __init__(
        self,
        agents: List[AgentBase],
        command_accumulation_policy: CommandAccumulatorBase,
    ):
        """Constructor.

        Args:
            agents (List[AgentBase]): List of Agents to be managed by this manager.
            command_accumulation_policy (CommandAccumulatorBase): The command accumulation
                policy to be used on the control outputs.
        """
        self.agents = agents
        self._accumulator = command_accumulation_policy

    def add_agent(self, agent: AgentBase):
        """Add an agent to be managed by this controller manager.

        Args:
            agent (AgentBase): The agent to be added to the end of the list.

        Returns:
            ControllerManagerBase: The self object with the new agent added.
                This is for using this method in a builder strategy.
        """
        assert isinstance(agent, AgentBase)
        self.agents.append(agent)
        return self

    def set_cmd_accumulation_policy(self, policy: CommandAccumulatorBase):
        """Set/change the command accumulation policy for this mananger.

        Args:
            policy (CommandAccumulatorBase): The new policy to be used.

        Returns:
            ControllerManagerBase: The self object with the policy updated.
                This is for using this method in a builder strategy.
        """
        assert isinstance(policy, CommandAccumulatorBase)
        self._accumulator = policy
        return self

    @abstractmethod
    def update(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> RobotCmd:
        """Run one update step of all the 'agents' managed by this manager.

        Args:
            robot_state (RobotState): Current robot state. This is equivalent to 'observation' in
                standard learning-based agents.
            global_plan (GlobalMotionPlan): The reference global plan to follow.
            t (float, optional): the current time signature of the control loop. Defaults to None
                (controllers may or may not need this).
            dt (float, optional): the time since the last control loop. Defaults to None
                (controllers may or may not need this).

        Returns:
            RobotCmd: The output control command (action) to be sent to the robot based on the
                accumulation policy used by this manager.

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.
        """
        raise NotImplementedError("This method has to be implemented in the child class")

    def shutdown(self):
        """Cleanly shutdown all agents managed by this controller manager."""
        for agent in self.agents:
            agent.shutdown()


# NOTE: Ideal controller manager should handle controller switching, starting, stopping etc.



"""A Python Robot Control Framework for quickly prototyping control algorithms for different robot
embodiments.

Primarily, this library provides an implementation of a typical control loop (via a
`MinimalCtrlLoop` (extended from `SimpleManagedCtrlLoop`) class), and defines interfaces for the
components in a control loop that can be used directly in these control loop implementations. It
also provides utility and debugging tools that will be useful for developing controllers and
planners for different robots. This package also provides implementations of basic controllers
and planners.

In the long run, this package will also provide implementations of popular motion planners and
controllers from literature and using existing libraries.
"""



from typing import List

from ...core.types import RobotState, Pose3D
from .robot_interface_base import RobotInterface
from ...utils.kinematics_dynamics import PinocchioInterface
from ...core.logging import throttled_logging


class RobotInterfaceWithPinocchio(RobotInterface):
    """
    Extension to the RobotInterface base class to provide easier interfacing with
    PinocchioInterface.

    NOTE: Any class inheriting this base class should call the constructor using
    ```
    super().__init__(
        robot_urdf=robot_urdf,
        ee_names=ee_names,
        floating_base=floating_base,
        verbose=verbose,
        create_pinocchio_interface=create_pinocchio_interface,
    )
    ```

    To update pinocchio in the control loop at every time, call
    `self.update_pinocchio_robot_state(robot_state=robot_state)` in the read() method
    implementation of the child class
    """

    def __init__(
        self,
        robot_urdf: str = None,
        ee_names: List[str] = (),
        floating_base: bool = False,
        verbose: bool = True,
        create_pinocchio_interface: bool = True,
    ):
        """Extension to the RobotInterface base class to provide easier interfacing with
        PinocchioInterface.

        NOTE: Any class inheriting this base class should call the constructor using
        ```
        super().__init__(
            robot_urdf=robot_urdf,
            ee_names=ee_names,
            floating_base=floating_base,
            verbose=verbose,
            create_pinocchio_interface=create_pinocchio_interface,
        )
        ```

        To update pinocchio in the control loop at every time, call
        `self.update_pinocchio_robot_state(robot_state=robot_state)` in the read() method
        implementation of the child class.


        Args:
            robot_urdf (str): Path to urdf file of robot.
            ee_names (List[str], optional): List of end-effectors for the robot. Defaults to None.
            floating_base (bool, optional): Specifying whether this robot has a floating or fixed
                base (setting to True creates a floating base instance with pinocchio, and makes
                the simulated robot free floating (not fixed base)). Defaults to True.
            verbose (bool, optional): Verbosity flag for debugging robot info during construction.
                Defaults to True.
            create_pinocchio_interface (bool, optional): If true, creates a PinocchioInterface
                object (mutable) for this robot.
        """
        self._pin_interface: PinocchioInterface = None
        self._urdf_path = robot_urdf
        if create_pinocchio_interface and robot_urdf is not None:
            if floating_base is None:
                raise ValueError(
                    f"{self.__class__.__name__}: Argument `floating_base` should be provided "
                    "to create PinocchioInterface."
                )
            self._pin_interface = PinocchioInterface(
                urdf_filename=robot_urdf,
                floating_base=floating_base,
                verbose=verbose,
                ee_names=ee_names,
            )

    def get_pinocchio_interface(self) -> PinocchioInterface:
        """Get the pinocchio interface associated with this robot.

        NOTE: this is a mutable object, and will be updated automatically
            whenever the read() method of this class instance is
            called.

        Returns:
            PinocchioInterface: A mutable copy of the PinocchioInterface object for
                this robot.
        """
        if self._pin_interface is None:
            raise ValueError(
                f"{self.__class__.__name__}: This robot was not initialised with a pinocchio interface! Set "
                "`robot_urdf=<path/to/combined_robot_urdf>` and `create_pinocchio_interface=True`"
                " during construction of object."
            )
        return self._pin_interface

    def update_pinocchio_robot_state(
        self, robot_state: RobotState, update_ee_states: bool = True
    ) -> RobotState:
        """Update the PinocchioInterface kinematics and dynamics data using the provided
        robot state, and return update robot state with info from pinocchio.

        Args:
            robot_state (RobotState): RobotState read from the robot.
            update_ee_states (bool, optional): If True, will populate
            `robot_state.state_estimates.end_effector_states` data using data from pinocchio.
            Defaults to True.

        Returns:
            RobotState: updated RobotState object.
        """
        if self._pin_interface is not None:
            self._pin_interface.update(
                global_base_position=robot_state.state_estimates.pose.position,
                global_base_quaternion=robot_state.state_estimates.pose.orientation,
                local_base_velocity_linear=robot_state.state_estimates.twist.linear,
                local_base_velocity_angular=robot_state.state_estimates.twist.angular,
                joint_positions=robot_state.joint_states.joint_positions,
                joint_velocities=robot_state.joint_states.joint_velocities,
                joint_order=robot_state.joint_states.joint_names,
            )
            if update_ee_states:
                robot_state.state_estimates.end_effector_states.ee_poses = [
                    Pose3D(*ee_pose) for ee_pose in self._pin_interface.get_ee_poses()
                ]
                robot_state.state_estimates.end_effector_states.ee_names = (
                    self._pin_interface.ee_names
                )
        else:
            throttled_logging.warning(
                f"{self.__class__.__name__}: Pinocchio interface does not "
                "exist for this robot. Not updating pinocchio or robot state.",
                delay_sec=2,
            )
        return robot_state

    @property
    def has_pinocchio_interface(self) -> bool:
        """Check if this robot has a usable pinocchio interface.

        Returns:
            bool: True if the robot interface has a PinocchioInterface.
        """
        return self._pin_interface is not None

    @property
    def urdf_path(self) -> str:
        """Path to the urdf used for creating this robot's pinocchio model."""
        return self._urdf_path



"""Visualise any robot urdf (or robot description package from robot_description.py) in pybullet.

Useful for finding out joint values and joint limits.

Usage: 
    pyrcf-visualise-robot <description_name> [as_floating_base? (1 or 0)]
    Args:
        <description_name>: Provide robot description name as first argument
            The robot description name should be a valid path to urdf file or 
            name of a description package from
            the `robot_descriptions.py` repo that has a valid urdf file: 
            https://github.com/robot-descriptions/robot_descriptions.py?tab=readme-ov-file#descriptions.
            e.g. `pepper_description`.
        (Optional): second argument can be 1 or 0 to load the robot as
            a fixed base robot (1) or not (0).
"""

from pyrcf.components.robot_interfaces.simulation import PybulletRobot
from pyrcf.utils.sim_utils import PybulletRobotVisualizer
from pyrcf.core.logging import logging
import sys
from pathlib import Path

usage_string = (
    "\nVisualise any robot urdf (or robot description package from robot_description.py) in pybullet.\n"
    "Useful for finding out joint values and joint limits.\n"
    "\n\nUsage: \n\n"
    "pyrcf-visualise-robot <description_name> [as_floating_base? (1 or 0)]\n\n"
    "Args:\n"
    "\t <description_name>: Provide robot description name as first argument\n"
    "\t\tThe robot description name should be a valid path to urdf file or \n"
    "\t\tname of a description package from\n"
    "\t\t the `robot_descriptions.py` repo that has a valid urdf file: \n"
    "\t\thttps://github.com/robot-descriptions/robot_descriptions.py?tab=readme-ov-file#descriptions.\n"
    "\t\t e.g. `pepper_description`.\n"
    "\t (Optional): second argument can be 1 or 0 to load the robot as\n"
    "\t\ta fixed base robot (1) or not (0).\n"
)


def main():

    if len(sys.argv) < 2:
        print(usage_string)
        return

    fixed_base: bool = True

    if len(sys.argv) > 2:
        fixed_base = bool(int(sys.argv[2]))

    logging.info(f"Robot has fixed base: {fixed_base}\n")

    try:
        if Path(sys.argv[1]).is_file():
            robot = PybulletRobot(
                urdf_path=sys.argv[1], floating_base=not fixed_base, enable_torque_mode=False
            )
        else:
            robot: PybulletRobot = PybulletRobot.fromAwesomeRobotDescriptions(
                robot_description_name=sys.argv[1],
                floating_base=not fixed_base,
                enable_torque_mode=False,
            )
    except Exception:
        print(usage_string)
        raise

    # load the visualiser using this robot
    viz: PybulletRobotVisualizer = PybulletRobotVisualizer.fromBulletRobot(
        pb_robot=robot,
        # optionally give starting joint positions (dictionary: joint_name -> joint_position)
        starting_joint_positions={},
        # optional list of strings to avoid when parsing through joints (eg. "_wheels")
        ignore_joints_with_str=[],
    )

    try:
        viz.run(sim_step_rate=240, slider_update_rate=10)
    except KeyboardInterrupt:
        viz.close()



from abc import abstractmethod

from ..pyrcf_component import PyRCFComponent
from ...core.types import GlobalMotionPlan, RobotState, LocalMotionPlan


class LocalPlannerBase(PyRCFComponent):
    """Base class defining the interface for all local planners that can be used in the control
    loop."""

    @abstractmethod
    def generate_local_plan(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> LocalMotionPlan:
        """The main local plan generation method to be called in the loop.

        This method will be called after the global plan is generated in the loop.

        This method should be implemented by any local planner implementation.

        Args:
            robot_state (RobotState): The current state information from the robot.
            global_plan (GlobalMotionPlan): the latest global plan generated by the global planner
                used in the loop.
            t (float): the current time signature of the control loop. Defaults to None
                (controllers may or may not need this).
            dt (float, optional): the time since the last control loop. Defaults to None
                (controllers may or may not need this).

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.

        Returns:
            LocalMotionPlan: The output local plan to be used by the controller in this iteration.
        """
        raise NotImplementedError("This method has to be implemented in the child class")


class DummyLocalPlanner(LocalPlannerBase):
    """Dummy local planner for testing pipeline."""

    def __init__(self, squawk: bool = True):
        """Dummy local planner for testing pipeline.

        Args:
            squawk (bool, optional): Verbose. Defaults to True.
        """
        self._squawk = squawk

    def generate_local_plan(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> LocalMotionPlan:
        """This is a dummy method for sanity checking the control loop."""
        if self._squawk:
            print(f"{__class__.__name__}: Generating dummy local plan. time: {t}, dt: {dt}")
        return LocalMotionPlan()



from abc import abstractmethod

from ..pyrcf_component import PyRCFComponent
from ...core.types import GlobalMotionPlan, RobotState


class GlobalMotionPlannerBase(PyRCFComponent):
    """Simple protocol and base class for a global motion planner interface.

    This protocol has to be respected by all classes that generate a GlobalMotionPlan object
    to be used in the control loop by local planners. (e.g. user interfaces such as
    keyboard_interface).

    INFO: A general 'global motion planner' is a planner that performs high-level planning that is
        dependent on the task the robot has to perform. For e.g. for a A->B locomotion, the global
        planner might generate a sequence of collision-aware waypoints for the local planner to
        respect.
    """

    @abstractmethod
    def generate_global_plan(
        self, robot_state: RobotState, t: float, dt: float
    ) -> GlobalMotionPlan:
        """Generate a global motion plan message.

        Args:
            robot_state (RobotState): Current robot state.
            t (float, optional): the current time signature of the control loop.
            dt (float, optional): the time since the last control loop.

        Returns:
            GlobalMotionPlan: Output global plan to be sent to local planner in the control loop.
        """



from abc import abstractmethod

from ..pyrcf_component import PyRCFComponent
from ...core.types import RobotState, RobotCmd, LocalMotionPlan


class ControllerBase(PyRCFComponent):
    """Base class defining the interface for all controllers that can be used in the control loop."""

    @abstractmethod
    def update(
        self,
        robot_state: RobotState,
        local_plan: LocalMotionPlan,
        t: float = None,
        dt: float = None,
    ) -> RobotCmd:
        """The main control update method to be called in the loop.

        This method will be called after the global plan and local plan are generated in the loop.

        This method should be implemented by any controller implementation.

        Args:
            robot_state (RobotState): The current state information from the robot.
            local_plan (LocalMotionPlan): the latest local plan generated by the local planner used
                in the loop.
            t (float, optional): the current time signature of the control loop. Defaults to None
                (controllers may or may not need this).
            dt (float, optional): the time since the last control loop. Defaults to None
                (controllers may or may not need this).

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.

        Returns:
            RobotCmd: The output control command to be sent to the robot.
        """
        raise NotImplementedError("This method has to be implemented in the child class")


class DummyController(ControllerBase):
    """Dummy controller for testing pipeline."""

    def __init__(self, squawk: bool = True):
        """Dummy controller for testing pipeline.

        Args:
            squawk (bool, optional): Verbosity. Defaults to True.
        """
        self._squawk = squawk

    def update(
        self,
        robot_state: RobotState,
        local_plan: LocalMotionPlan,
        t: float = None,
        dt: float = None,
    ) -> RobotCmd:
        """This is a dummy method for sanity checking the control loop."""
        if self._squawk:
            print(f"{__class__.__name__}: Performing dummy control update. time: {t}, dt: {dt}")
        return RobotCmd()



from typing import List, Tuple

from ..core.logging import logging
from ..core.exceptions import CtrlLoopExitSignal
from ..components.robot_interfaces.robot_interface_base import RobotInterface
from ..components.controller_manager.controller_manager_base import ControllerManagerBase
from ..components.global_planners.global_planner_base import GlobalMotionPlannerBase
from ..components.state_estimators.state_estimator_base import StateEstimatorBase
from ..components.ctrl_loop_debuggers.ctrl_loop_debugger_base import CtrlLoopDebuggerBase
from ..components.callback_handlers.base_callbacks import CustomCallbackBase
from ..utils.time_utils import RateLimiter, PythonPerfClock, ClockBase


class _True_dt:

    def __init__(self):
        self._clock = PythonPerfClock()
        self._prev_t = self._clock.get_time()

    def get_dt(self):
        curr_t = self._clock.get_time()
        dt = curr_t - self._prev_t
        self._prev_t = curr_t
        return dt


class SimpleManagedCtrlLoop:
    """A minimal control loop that takes in a controller manager instead of a single controller and
    local planner.

    The run method runs the following in a loop of specified frequency.

    1. read robot state from robot interface
    2. update robot state with state estimates from state estimator
    3. get global plan message from global planner / user interface
    4. for agent in controller_manager.agents:
        a. compute control command to track the global plan (agent may be a Planner+Controller
            agent)
    5. accumulate commands from accumulation policy used by controller manager
    6. write control command to robot
    """

    def __init__(
        self,
        robot_interface: RobotInterface,
        state_estimator: StateEstimatorBase,
        controller_manager: ControllerManagerBase,
        global_planner: GlobalMotionPlannerBase,
        verbose: bool = True,
    ):
        """Create a minimal control loop where agents (controller+localplanner) are run in sequence.

        Args:
            robot_interface (RobotInterface): The robot interface to use in the loop.
            state_estimator (StateEstimatorBase): State estimator object to use in the loop.
            controller_manager (ControllerManagerBase): The controller manager object with agents
                pre-loaded.
            global_planner (GlobalMotionPlannerBase): The global planner object to be used in the
                loop.
        """
        self.robot = robot_interface
        self.state_estimator = state_estimator
        self.controller_manager = controller_manager
        self.global_planner = global_planner

        self._true_dt_getter = _True_dt()
        self._true_dt = 0.0
        self._loop_count = 0

        if verbose:
            msg = "\n----------------------------------------------------------------------------\n"
            msg += f"{self.__class__.__name__}: Control loop components:\n"
            msg += f"\tRobot: {self.robot.get_class_info()}\n"
            msg += f"\tGlobal Planner: {self.global_planner.get_class_info()}\n"
            msg += f"\tState Estimator: {self.state_estimator.get_class_info()}\n"
            msg += "\tAgents:\n"
            for n, agent in enumerate(self.controller_manager.agents):
                msg += f"\t\tAgent {n+1}: {agent.get_class_info()}\n"
            msg += "----------------------------------------------------------------------------\n"
            logging.info(msg=msg)

    def run(
        self,
        loop_rate: float,
        clock: ClockBase = PythonPerfClock(),
        debuggers: List[CtrlLoopDebuggerBase] = None,
        prestep_callbacks: List[CustomCallbackBase] = None,
        poststep_callbacks: List[CustomCallbackBase] = None,
    ):
        """The main control loop.

        Args:
            loop_rate (float): loop rate in hz. This will always use the system clock to try and
                maintain the desired control loop rate.
            clock (ClockBase, optional): Defines the clock to use for querying time for the
                components in the control loop. Time will be queried once per loop and the same `t`
                and `dt` will be passed to all components in that iteration. The option to select
                the clock is so that sim time can be used if required.
            debuggers (CtrlLoopDebuggerBase, optional): List of `CtrlLoopDebuggerBase` objects that
                are to be run in the control loop.
            prestep_callbacks (List[CustomCallbackBase], optional): List of `CustomCallbackBase` objects to
                be run in the control loop. Not recommended to do this except for debugging.
                Defaults to None.
            poststep_callbacks (List[CustomCallbackBase]. optional): List of additional `CustomCallbackBase`
                objects to be run in the control loop. Not recommended to do this except for
                debugging. Defaults to None.
        """
        if prestep_callbacks is None:
            prestep_callbacks = []

        if poststep_callbacks is None:
            poststep_callbacks = []

        for cb in prestep_callbacks + poststep_callbacks:
            assert isinstance(cb, CustomCallbackBase)

        logging.info(f"{self.__class__.__name__}: Activating robot...")
        while not self.robot.activate():
            continue
        logging.info(f"{self.__class__.__name__}: Robot activated.")

        rate = RateLimiter(frequency=loop_rate, name="control loop", clock=PythonPerfClock())
        start_t: float = clock.get_time()
        prev_t: float = clock.get_time() - start_t
        while True:
            try:
                curr_t: float = clock.get_time() - start_t
                dt: float = curr_t - prev_t

                self._loop_count += 1
                self._true_dt = self._true_dt_getter.get_dt()

                for cb in prestep_callbacks:
                    cb.run_once()

                # read latest robot state
                robot_state = self.robot.read()

                # use state estimator to compute robot states that are not directly observable from
                # the robot interface such as robot pose in the world, base velocity, foot contact
                # states, etc.
                robot_state = self.state_estimator.update_robot_state_with_state_estimates(
                    robot_state=robot_state, t=curr_t, dt=dt
                )

                # generate global plan for local planner
                global_plan = self.global_planner.generate_global_plan(
                    robot_state=robot_state, t=curr_t, dt=dt
                )

                # use the local plan to generate instantaneous command to be sent to the robot
                cmd = self.controller_manager.update(
                    robot_state=robot_state, global_plan=global_plan, t=curr_t, dt=dt
                )

                # write the commands to the robot
                self.robot.write(cmd=cmd)

                if debuggers is not None:
                    agent_outputs = [
                        agent.get_last_output() for agent in self.controller_manager.agents
                    ]
                    for debugger in debuggers:
                        debugger.run_once(
                            t=curr_t,
                            dt=dt,
                            robot_state=robot_state,
                            global_plan=global_plan,
                            agent_outputs=agent_outputs,
                            robot_cmd=cmd,
                        )

                for cb in poststep_callbacks:
                    cb.run_once()

                prev_t = curr_t

                rate.sleep()
            except (KeyboardInterrupt, CtrlLoopExitSignal):
                logging.info(
                    "Received control loop exit signal. Attempting shutdown of components..."
                )
                logging.info(f"{self.__class__.__name__}: Deactivating robot...")
                while not self.robot.deactivate():
                    continue
                logging.info(f"{self.__class__.__name__}: Robot deactivated.")
                self.shutdown()
                break

        if debuggers is not None:
            for debugger in debuggers:
                debugger.shutdown()

        for cb in prestep_callbacks + poststep_callbacks:
            cb.cleanup()

    def get_actual_loop_rate(self) -> Tuple[float, float]:
        """Get the true loop rate and dt (using system clock).

        Returns:
            Tuple[float, float]: True dt (s), true loop rate (hz)
        """
        return self._true_dt, (1 / self._true_dt) if self._true_dt > 0 else 0

    def get_loop_count(self) -> int:
        """Get the current number of iterations of the control loop.

        Returns:
            int: current number of iterations of the control loop
        """
        return self._loop_count

    def shutdown(self):
        """Cleanly shutdown all components in the control loop.
        Sequence of shutdown:
            - robot interface
            - controller manager
            - global planner
        """
        logging.info(f"{self.__class__.__name__}: Shutting down.")
        self.robot.shutdown()
        self.controller_manager.shutdown()
        self.global_planner.shutdown()
        logging.info(f"{self.__class__.__name__}: Control loop shut down complete.")

    def send_shutdown_signal(self):
        """Request clean shutdown of control loop and components."""
        raise CtrlLoopExitSignal



from abc import abstractmethod

from ..pyrcf_component import PyRCFComponent
from ...core.types import RobotState, RobotCmd


class RobotInterface(PyRCFComponent):
    """Base class defining the interface for all robots that can be used in the control loop."""

    @abstractmethod
    def read(self) -> RobotState:
        """Get the latest state from the robot.

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.

        Returns:
            RobotState: the latest state of the robot.
        """
        raise NotImplementedError("This method has to be implemented in the child class")

    @abstractmethod
    def write(self, cmd: RobotCmd) -> bool:
        """Send the specified commands to the robot.

        Args:
            cmd (RobotCmd): the command object as computed by the controller in the loop.

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.

        Returns:
            bool: True if sending command was succesful.
        """
        raise NotImplementedError("This method has to be implemented in the child class")

    def activate(self) -> bool:
        """This method will be called in a loop until it returns true, before the control loop
        starts. Override in child class as required."""
        return True

    def deactivate(self) -> bool:
        """This method will be called in a loop until it returns true, after the control loop is
        stopped, and before shutdown of this class is called. Override in child class as required.
        """
        return True


class DummyRobot(RobotInterface):
    def __init__(self, squawk: bool = True):
        """Dummy robot class for testing pipeline.

        Args:
            squawk (bool, optional): Verbose. Defaults to True.
        """
        self._squawk = squawk

    def read(self) -> RobotState:
        """This is a dummy method for sanity checking the control loop."""
        if self._squawk:
            print(f"{__class__.__name__}: Reading state from dummy robot")
        return RobotState()

    def write(self, cmd: RobotCmd) -> bool:
        """This is a dummy method for sanity checking the control loop."""
        if self._squawk:
            print(f"{__class__.__name__}: Writing cmd to dummy robot")
        return True



from typing import TypeAlias, Tuple
from dataclasses import dataclass, field
import numpy as np

QuatType: TypeAlias = np.ndarray  # TODO: maybe use quaternionic package here
"""Numpy array representating quaternion in format [x,y,z,w]"""

Vector3D: TypeAlias = np.ndarray
"""Numpy array representating 3D cartesian vector in format [x,y,z]"""


@dataclass
class Pose3D:
    """Container for holding 3D pose values.

    This object also allows equality comparison (e.g. check for (`if pose1 == pose2:`)).

    NOTE: When creating this object, use `validate_quaternion=True` to assert quaternion validity.
        Defaults to False.
    """

    position: Vector3D = field(default_factory=lambda: np.zeros(3))
    orientation: QuatType = field(default_factory=lambda: np.array([0, 0, 0, 1]))
    """quaternion [x,y,z,w]"""

    validate_quaternion: bool = field(default=False, init=True, repr=False)
    """If this is enabled, quaternion validity will be checked after construction. Default False."""

    def __eq__(self, other: "Pose3D") -> bool:
        return np.allclose(self.position, other.position) and np.allclose(
            self.orientation, other.orientation
        )

    def __setattr__(self, prop, val):
        match prop:
            case "position":
                assert len(val) == 3, "Position dimension should be 3."
            case "orientation":
                assert len(val) == 4, "Quaternion dimension should be 4."
                if self.validate_quaternion:
                    assert np.isclose(np.linalg.norm(val), 1.0), "Quaternion is not normalised."
            case "validate_quaternion":
                super().__setattr__(prop, val)
                return
            case _:
                raise NameError(
                    f"Tried to assign value to illegal parameter {prop} in object of type "
                    f"{self.__class__.__name__}."
                )
        super().__setattr__(prop, np.array(val))


@dataclass
class Twist:
    """Container for holding velocity in the base frame of the robot.

    This object also allows equality comparison (e.g. check for (`if twist1 == twist2:`)).
    """

    linear: Vector3D = field(default_factory=lambda: np.zeros(3))
    angular: Vector3D = field(default_factory=lambda: np.zeros(3))

    def __eq__(self, other: "Twist") -> bool:
        return np.allclose(self.linear, other.linear) and np.allclose(self.angular, other.angular)

    def __setattr__(self, __name: str, __value: np.ndarray | Tuple[float]) -> None:
        if __name not in ["linear", "angular"]:
            raise NameError(
                f"Tried to assign value to illegal parameter {__name} in object of type "
                f"{self.__class__.__name__}."
            )
        assert len(__value) == 3, f"{__name} velocity should be of dimension 3"
        super().__setattr__(__name, np.array(__value))


@dataclass
class RelativePose:
    """This is used for representing relative pose (typically desired body pose commands) where RPY
    is better.

    This object also allows equality comparison (e.g. check for (`if r_pose1 == r_pose2:`)).
    """

    position: Vector3D = field(default_factory=lambda: np.zeros(3))
    rpy: Vector3D = field(default_factory=lambda: np.zeros(3))
    """[Roll, Pitch, Yaw] in radians"""

    def __eq__(self, other: "RelativePose") -> bool:
        return np.allclose(self.position, other.position) and np.allclose(self.rpy, other.rpy)

    def __setattr__(self, __name: str, __value: np.ndarray | Tuple[float]) -> None:
        if __name not in ["position", "rpy"]:
            raise NameError(
                f"Tried to assign value to illegal parameter {__name} in object of type "
                f"{self.__class__.__name__}."
            )
        assert len(__value) == 3, f"{__name} should be of dimension 3"
        super().__setattr__(__name, np.array(__value))
